"""Contains a class to implement mode devices."""
from typing import Optional

import abc

from mpf.core.device import Device
from mpf.core.events import event_handler
from mpf.core.machine import MachineController
from mpf.core.mode import Mode
from mpf.core.player import Player


class ModeDevice(Device, metaclass=abc.ABCMeta):

    """A device in a mode."""

    __slots__ = ["mode"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialise mode device."""
        super().__init__(machine, name)
        self.mode = None    # type: Optional[Mode]

    async def device_added_to_mode(self, mode: Mode) -> None:
        """Add device to a running mode.

        Args:
        ----
            mode: Mode which loaded the device
        """
        del mode
        await self._initialize()

    def device_loaded_in_mode(self, mode: Mode, player: Player) -> None:
        """Load device in running mode.

        The mode just started.

        Args:
        ----
            mode: Mode which loaded the device
            player: Current active player
        """
        del player
        self.mode = mode

    @property
    def can_exist_outside_of_game(self) -> bool:
        """Return true if this device can exist outside of a game."""
        return False

    def overload_config_in_mode(self, mode: Mode, config: dict) -> None:
        """Overload config in mode."""
        del mode
        del config
        raise AssertionError("Device {} cannot be overloaded.".format(self))

    @event_handler(20)
    def event_enable(self, **kwargs):
        """Event handler for enable event."""
        del kwargs
        self.enable()

    def enable(self) -> None:
        """Enable handler."""

    def add_control_events_in_mode(self, mode: Mode) -> None:
        """Add control events in mode if this device has any mode control events.

        Args:
        ----
            mode: Mode which loaded the device
        """
        if "enable_events" in self.config and not self.config['enable_events']:
            mode.add_mode_event_handler("mode_{}_started".format(mode.name),
                                        self.event_enable, priority=100)

    def remove_control_events_in_mode(self) -> None:
        """Remove control events."""

    def device_removed_from_mode(self, mode: Mode) -> None:
        """Remove device because mode is unloading.

        Device object will continue to exist and may be added to the mode again later.

        Args:
        ----
            mode: Mode which stopped
        """
        del mode
        self.mode = None



"""Contains the base classes for drop targets and drop target banks."""
from enum import Enum
from typing import List, Optional
from typing import Set

from mpf.core.mode import Mode
from mpf.core.player import Player

from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.mode_device import ModeDevice
from mpf.core.system_wide_device import SystemWideDevice

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.devices.driver import Driver           # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor("complete")
class DropTarget(SystemWideDevice):

    """Represents a single drop target in a pinball machine.

    Args: Same as the `Target` parent class
    """

    config_section = 'drop_targets'
    collection = 'drop_targets'
    class_label = 'drop_target'

    __slots__ = ["reset_coil", "knockdown_coil", "banks", "_in_ball_search", "complete", "delay", "_ignore_switch_hits"]

    def __init__(self, machine: "MachineController", name: str) -> None:
        """Initialise drop target."""
        self.reset_coil = None              # type: Optional[Driver]
        self.knockdown_coil = None          # type: Optional[Driver]
        self.banks = set()                  # type: Set[DropTargetBank]
        super().__init__(machine, name)

        self._in_ball_search = False
        self.complete = False
        self.delay = DelayManager(machine)

        self._ignore_switch_hits = False

    async def _initialize(self):
        await super()._initialize()
        self.reset_coil = self.config['reset_coil']
        self.knockdown_coil = self.config['knockdown_coil']

        # can't read the switch until the switch controller is set up
        self.machine.events.add_handler('init_phase_4',
                                        self._update_state_from_switch, priority=2)
        self.machine.events.add_handler('init_phase_4',
                                        self._register_switch_handlers, priority=1)

        if self.config['ball_search_order']:
            self.config['playfield'].ball_search.register(
                self.config['ball_search_order'], self._ball_search, self.name)

        if '{}_active'.format(self.config['playfield'].name) in self.config['switch'].tags:
            self.raise_config_error(
                "Drop target device '{}' uses switch '{}' which has a "
                "'{}_active' tag. This is handled internally by the device. Remove the "
                "redundant '{}_active' tag from that switch.".format(
                    self.name, self.config['switch'].name, self.config['playfield'].name,
                    self.config['playfield'].name), 1)

    def _ignore_switch_hits_for(self, ms):
        """Ignore switch hits for ms."""
        self._ignore_switch_hits = True
        self.delay.reset(name="ignore_switch", callback=self._restore_switch_hits, ms=ms)

    def _restore_switch_hits(self):
        self._ignore_switch_hits = False
        self._update_state_from_switch(reconcile=True)

    def _ball_search_phase1(self):
        if not self.complete and self.reset_coil:
            self.reset_coil.pulse()
            return True
        # if down. knock down again
        if self.complete and self.knockdown_coil:
            self.knockdown_coil.pulse()
            return True
        return False

    def _ball_search_phase2(self):
        if self.reset_coil and self.knockdown_coil:
            self._in_ball_search = True
            if self.complete:
                self.reset_coil.pulse()
                self.delay.add(100, self._ball_search_knockdown)
            else:
                self.knockdown_coil.pulse()
                self.delay.add(100, self._ball_search_reset)
            return True

        # fall back to phase1
        return self._ball_search_phase1()

    def _ball_search_phase3(self):
        if self.complete:
            if self.reset_coil:
                self.reset_coil.pulse()
                if self.knockdown_coil:
                    self._in_ball_search = True
                    self.delay.add(100, self._ball_search_knockdown)
                return True

            # fall back to phase1
            return self._ball_search_phase1()

        if self.knockdown_coil:
            self.knockdown_coil.pulse()
            if self.reset_coil:
                self._in_ball_search = True
                self.delay.add(100, self._ball_search_reset)
            return True

        # fall back to phase1
        return self._ball_search_phase1()

    def _ball_search_iteration_finish(self):
        self._in_ball_search = False

    def _ball_search_knockdown(self):
        self.knockdown_coil.pulse()
        self.delay.add(100, self._ball_search_iteration_finish)

    def _ball_search_reset(self):
        self.reset_coil.pulse()
        self.delay.add(100, self._ball_search_iteration_finish)

    def _ball_search(self, phase, iteration):
        del iteration
        if phase == 1:
            # phase 1: do not change state.
            # if up. reset again
            return self._ball_search_phase1()
        if phase == 2:
            # phase 2: if we can reset and knockdown the target we will do that
            return self._ball_search_phase2()

        # phase3: reset no matter what
        return self._ball_search_phase3()

    def _register_switch_handlers(self, **kwargs):
        del kwargs
        # register for notification of switch state
        # this is in addition to the parent since drop targets track
        # self.complete in separately

        self.machine.switch_controller.add_switch_handler_obj(
            self.config['switch'],
            self._update_state_from_switch, 0)
        self.machine.switch_controller.add_switch_handler_obj(
            self.config['switch'],
            self._update_state_from_switch, 1)

    @event_handler(6)
    def event_enable_keep_up(self, **kwargs):
        """Handle enable_keep_up control event."""
        del kwargs
        self.enable_keep_up()

    def enable_keep_up(self):
        """Keep the target up by enabling the coil."""
        if self.reset_coil:
            self.reset_coil.enable()

    @event_handler(5)
    def event_disable_keep_up(self, **kwargs):
        """Handle disable_keep_up control event."""
        del kwargs
        self.disable_keep_up()

    def disable_keep_up(self):
        """No longer keep up the target up."""
        if self.reset_coil:
            self.reset_coil.disable()

    @event_handler(7)
    def event_knockdown(self, **kwargs):
        """Handle knockdown control event."""
        del kwargs
        self.knockdown()

    def knockdown(self):
        """Pulse the knockdown coil to knock down this drop target."""
        if self.knockdown_coil and not self.machine.switch_controller.is_active(self.config['switch']):
            self._ignore_switch_hits_for(ms=self.config['ignore_switch_ms'])
            self.knockdown_coil.pulse(max_wait_ms=self.config['knockdown_coil_max_wait_ms'])

    def _update_state_from_switch(self, reconcile=False, **kwargs):
        del kwargs

        is_complete = self.machine.switch_controller.is_active(
            self.config['switch'])

        if self._in_ball_search or self._ignore_switch_hits:
            return

        if not reconcile:
            self.config['playfield'].mark_playfield_active_from_device_action()

        if is_complete != self.complete:

            if is_complete:
                self._down()
            else:
                self._up()

            self._update_banks()

    def _down(self):
        self.complete = True
        self.machine.events.post('drop_target_' + self.name + '_down', device=self)
        '''event: drop_target_(name)_down
        desc: The drop target with the (name) has just changed to the "down"
        state.'''

    def _up(self):
        self.complete = False
        self.machine.events.post('drop_target_' + self.name + '_up', device=self)
        '''event: drop_target_(name)_up
        desc: The drop target (name) has just changed to the "up" state.'''

    def _update_banks(self):
        for bank in self.banks:
            bank.member_target_change()

    def add_to_bank(self, bank):
        """Add this drop target to a drop target bank.

         This allows the bank to update its status based on state changes to this drop target.

        Args:
        ----
            bank: DropTargetBank object to add this drop target to.
        """
        self.banks.add(bank)

    def external_reset_from_bank(self):
        """Handle the reset from our bank.

        The bank might pulse the coil from this device or it might have a
        separate reset coil which will trigger a reset on switch of this
        device.
        Make sure we do not mark the playfield as active.
        """
        if self.machine.switch_controller.is_active(self.config['switch']):
            self._ignore_switch_hits_for(ms=self.config['ignore_switch_ms'])

    def remove_from_bank(self, bank):
        """Remove the DropTarget from a bank.

        Args:
        ----
            bank: DropTargetBank object to remove
        """
        self.banks.remove(bank)

    @event_handler(1)
    def event_reset(self, **kwargs):
        """Handle reset control event."""
        del kwargs
        self.reset()

    def reset(self):
        """Reset this drop target.

        If this drop target is configured with a reset coil, then this method
        will pulse that coil. If not, then it checks to see if this drop target
        is part of a drop target bank, and if so, it calls the reset() method of
        the drop target bank.

        This method does not reset the target profile, however, the switch event
        handler should reset the target profile on its own when the drop target
        physically moves back to the up position.
        """
        if self.reset_coil and self.machine.switch_controller.is_active(self.config['switch']):
            self._ignore_switch_hits_for(ms=self.config['ignore_switch_ms'])
            self.reset_coil.pulse(max_wait_ms=self.config['reset_coil_max_wait_ms'])


class DropTargetBankState(Enum):

    """States of the drop target bank."""

    UNKNOWN = "unknown"
    UP = "up"
    DOWN = "down"
    MIXED = "mixed"


@DeviceMonitor("complete", "down", "up", "state")
class DropTargetBank(SystemWideDevice, ModeDevice):

    """A bank of drop targets in a pinball machine by grouping together multiple `DropTarget` class devices."""

    config_section = 'drop_target_banks'
    collection = 'drop_target_banks'
    class_label = 'drop_target_bank'

    def __init__(self, machine: "MachineController", name: str) -> None:
        """Initialise drop target bank."""
        super().__init__(machine, name)

        self.drop_targets = list()          # type: List[DropTarget]
        self.reset_coil = None              # type: Optional[Driver]
        self.reset_coils = set()            # type: Set[Driver]
        self.complete = False
        self.state = DropTargetBankState.UNKNOWN
        self.down = 0
        self.up = 0
        self.delay = DelayManager(machine)
        self._ignore_switch_hits = False

    @property
    def can_exist_outside_of_game(self):
        """Return true if this device can exist outside of a game."""
        return True

    async def _initialize(self):
        await super()._initialize()
        self.drop_targets = self.config['drop_targets']
        self.reset_coil = self.config['reset_coil']
        self.reset_coils = self.config['reset_coils']

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Add targets."""
        self._add_targets_to_bank()

    async def device_added_system_wide(self):
        """Add targets."""
        await super().device_added_system_wide()
        self._add_targets_to_bank()

    def _add_targets_to_bank(self):
        """Add targets to bank."""
        for target in self.drop_targets:
            target.add_to_bank(self)

        self.member_target_change()

        self.debug_log('Drop Targets: %s', self.drop_targets)

    @event_handler(5)
    def event_reset(self, **kwargs):
        """Handle reset control event."""
        del kwargs
        self.reset()

    def reset(self):
        """Reset this bank of drop targets.

        This method has some intelligence to figure out what coil(s) it should
        fire. It builds up a set by looking at its own reset_coil and
        reset_coils settings, and also scanning through all the member drop
        targets and collecting their coils. Then it pulses each of them. (This
        coil list is a "set" which means it only sends a single pulse to each
        coil, even if each drop target is configured with its own coil.)
        """
        self.debug_log('Resetting')

        if self.down == 0:
            self.info_log('All targets are already up. Will not reset bank.')
            return

        self.info_log('%s targets are down. Will reset those.', self.down)

        # figure out all the coils we need to pulse
        coils = set()       # type: Set[Driver]

        for drop_target in self.drop_targets:
            # add all reset coil for targets which are down
            if drop_target.reset_coil and drop_target.complete:
                coils.add(drop_target.reset_coil)

            # tell the drop target that we are going to reset it physically
            drop_target.external_reset_from_bank()

        for coil in self.reset_coils:
            coils.add(coil)

        if self.reset_coil:
            coils.add(self.reset_coil)

        if self.config['ignore_switch_ms']:
            self._ignore_switch_hits = True
            self.delay.add(ms=self.config['ignore_switch_ms'],
                           callback=self._restore_switch_hits,
                           name='ignore_hits')

        # now pulse them
        for coil in coils:
            self.debug_log('Pulsing reset coils: %s', coils)
            coil.pulse(max_wait_ms=self.config['reset_coil_max_wait_ms'])

    def _restore_switch_hits(self):
        self._ignore_switch_hits = False
        self.member_target_change()

    def member_target_change(self):
        """Handle that a member drop target has changed state.

        This method causes this group to update its down and up counts and
        complete status.
        """
        if self._ignore_switch_hits:
            return

        self.down = 0
        self.up = 0

        for target in self.drop_targets:
            if target.complete:
                self.down += 1
            else:
                self.up += 1

        self.debug_log(
            'Member drop target status change: Up: %s, Down: %s,'
            ' Total: %s', self.up, self.down,
            len(self.drop_targets))

        if self.down == len(self.drop_targets):
            self._bank_down()
        elif not self.down:
            self._bank_up()
        else:
            self._bank_mixed()

    def _bank_down(self):
        if self.state == DropTargetBankState.DOWN:
            return
        self.state = DropTargetBankState.DOWN
        self.complete = True
        self.debug_log('All targets are down')

        if self.config['reset_on_complete']:
            self.debug_log("Reset on complete after %s", self.config['reset_on_complete'])
            self.delay.add(self.config['reset_on_complete'], self.reset)

        self.machine.events.post('drop_target_bank_' + self.name + '_down')
        '''event: drop_target_bank_(name)_down
        desc: Every drop target in the drop target bank called
        (name) is now in the "down" state. This event is
        only posted once, when all the drop targets are down.'''

    def _bank_up(self):
        if self.state == DropTargetBankState.UP:
            return
        self.state = DropTargetBankState.UP
        self.complete = False
        self.debug_log('All targets are up')
        self.machine.events.post('drop_target_bank_' + self.name + '_up')
        '''event: drop_target_bank_(name)_up
        desc: Every drop target in the drop target bank called
        (name) is now in the "up" state. This event is
        only posted once, when all the drop targets are up.'''

    def _bank_mixed(self):
        if self.state == DropTargetBankState.MIXED:
            return
        self.state = DropTargetBankState.MIXED
        self.complete = False
        self.machine.events.post('drop_target_bank_' + self.name + '_mixed',
                                 down=self.down)
        '''event: drop_target_bank_(name)_mixed
        desc: The drop targets in the drop target bank
        (name) are in a "mixed" state, meaning that they're
        not all down or not all up. This event is posted every time a member
        drop target changes but the overall bank is not not complete.'''

    def device_removed_from_mode(self, mode):
        """Remove targets which were added in this mode."""
        for target in self.drop_targets:
            target.remove_from_bank(self)



# All the individual config players have their own tests, but this tests the
# generic ConfigPlayer functionality
from collections import namedtuple

from mpf.assets.show import Show
from mpf.config_players.device_config_player import DeviceConfigPlayer
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.core.config_player import ConfigPlayer

PlayCall = namedtuple('PlayCall', ['settings', 'key', 'priority', 'kwargs'])


class BananaPlayer(DeviceConfigPlayer):
    config_file_section = 'banana_player'
    show_section = 'bananas'
    machine_collection_name = 'bananas'

    def __init__(self, machine):
        super().__init__(machine)
        self.machine.bananas = dict()
        self.machine.banana_play_calls = list()

    def play(self, settings, context, calling_context, key=None, priority=0, start_time=None, **kwargs):
        del start_time
        self.machine.banana_play_calls.append(PlayCall(
            settings, key, priority, kwargs))

    def _clear(self, key):
        pass

    def get_express_config(self, value):
        return dict(banana=value)

    def get_full_config(self, value):
        return value


class TestConfigPlayers(MpfTestCase):

    def get_config_file(self):
        return 'test_config_players.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/config_players/'

    def setUp(self):
        self.machine_config_patches['mpf']['config_players'] = dict()
        self.machine_config_patches['mpf']['config_players']['banana'] = \
            'mpf.tests.test_ConfigPlayers.BananaPlayer'

        self.machine_spec_patches['banana_player'] = dict(__valid_in__='machine, mode')

        # Hack around globals in shows
        Show.next_id = 0

        super().setUp()

    def test_config_player(self):
        self.assertIn('bananas', self.machine.show_controller.show_players)

        # post events to make sure banana_player is called
        self.machine.events.post('event1')
        self.advance_time_and_run()

        play_call = self.machine.banana_play_calls.pop()

        self.assertEqual(play_call.settings, {'express': {}})
        self.assertEqual(play_call.key, None)
        self.assertEqual(play_call.kwargs, {})

        self.machine.events.post('event2')
        self.advance_time_and_run()

        play_call = self.machine.banana_play_calls.pop()

        self.assertEqual(play_call.settings,
                         {'some': {'banana': 'key'}})
        self.assertEqual(play_call.key, None)
        self.assertEqual(play_call.kwargs, {})  # todo

        self.machine.events.post('event3')
        self.advance_time_and_run()

        play_call = self.machine.banana_play_calls.pop()

        self.assertEqual(play_call.settings,
                         {'this_banana': {'some': 'key'},
                          'that_banana': {'some': 'key'}})
        self.assertEqual(play_call.key, None)
        self.assertEqual(play_call.kwargs, {})  # todo

        # event5 is in mode1, so make sure it is not called now

        self.assertEqual(0, len(self.machine.banana_play_calls))

        self.machine.events.post('event5')
        self.advance_time_and_run()

        self.assertEqual(0, len(self.machine.banana_play_calls))

        # Start the mode, make sure the mode player enables
        self.machine.modes['mode1'].start()
        self.advance_time_and_run()

        self.machine.events.post('event5')
        self.advance_time_and_run()

        play_call = self.machine.banana_play_calls.pop()

        self.assertEqual(play_call.settings, {'express': {}})
        # Mode should be passed properly
        # self.assertEqual(play_call.key, 'mode1')
        self.assertEqual(play_call.kwargs, {})  # todo

        # stop the mode, make sure the event doesn't fire
        self.machine.modes['mode1'].stop()
        self.advance_time_and_run()

        self.machine.events.post('event5')
        self.advance_time_and_run()

        self.assertEqual(0, len(self.machine.banana_play_calls))

        # Start a show
        self.machine.events.post('event4')
        self.advance_time_and_run()

        play_call = self.machine.banana_play_calls.pop()
        self.assertEqual(play_call.settings, {'banana1': {'banana': 'express'}})
        # self.assertEqual(play_call.key, 'show1.1')
        self.assertEqual(play_call.kwargs, {'show_tokens': {}})  # todo

        self.assertEqual(1, len(self.machine.show_player.instances['_global']['show_player']))

        # todo add tests for mode 1 show, make sure the mode is passed
        # todo make sure it stops when the mode ends, that banana clear is
        # called when it stops, and that it doesn't start again once the mode
        # is not running

    def test_empty_config_player_section(self):
        self.machine.modes["mode2"].start()
        self.advance_time_and_run()
        self.machine.modes["mode2"].stop()
        self.advance_time_and_run()



from mpf.tests.MpfTestCase import MpfTestCase


class TestHead2Head(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/head2head/'

    def get_platform(self):
        return 'smart_virtual'

    def _prepare_troughs(self):
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_trough_front"])
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_trough_front"])
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_trough_front"])
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_trough_back"])
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_trough_back"])
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_trough_back"])

        self.advance_time_and_run()

        self.assertEqual(3, self.machine.ball_devices["bd_trough_front"].balls)
        self.assertEqual(3, self.machine.ball_devices["bd_trough_back"].balls)
        self.assertEqual(0, self.machine.playfields["playfield_front"].balls)
        self.assertEqual(0, self.machine.playfields["playfield_back"].balls)

    def test_loop(self):
        self._prepare_troughs()
        self.machine.ball_devices["bd_trough_front"].eject(target=self.machine.ball_devices["bd_trough_back"])
        self.machine.ball_devices["bd_trough_back"].eject(target=self.machine.ball_devices["bd_trough_front"])
        self.machine.ball_devices["bd_trough_front"].eject(target=self.machine.ball_devices["bd_trough_back"])
        self.machine.ball_devices["bd_trough_back"].eject(target=self.machine.ball_devices["bd_trough_front"])

        self.advance_time_and_run(100)

        self.assertEqual("idle", self.machine.ball_devices["bd_trough_front"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_trough_back"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_launcher_front"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_launcher_back"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_feeder_front"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_feeder_back"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_middle_front"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_middle_back"]._state)

        self.assertEqual(3, self.machine.ball_devices["bd_trough_front"].balls)
        self.assertEqual(3, self.machine.ball_devices["bd_trough_back"].balls)
        self.assertEqual(0, self.machine.playfields["playfield_front"].balls)
        self.assertEqual(0, self.machine.playfields["playfield_back"].balls)

    def testEject(self):
        self._prepare_troughs()
        self.machine.playfields["playfield_front"].add_ball()
        self.machine.playfields["playfield_back"].add_ball()

        self.assertEqual(2, self.machine.ball_devices["bd_trough_front"].available_balls)
        self.assertEqual(2, self.machine.ball_devices["bd_trough_back"].available_balls)
        self.assertEqual(1, self.machine.playfields["playfield_front"].available_balls)
        self.assertEqual(1, self.machine.playfields["playfield_back"].available_balls)
        self.assertEqual(3, self.machine.ball_devices["bd_trough_front"].balls)
        self.assertEqual(3, self.machine.ball_devices["bd_trough_back"].balls)
        self.assertEqual(0, self.machine.playfields["playfield_front"].balls)
        self.assertEqual(0, self.machine.playfields["playfield_back"].balls)

        self.advance_time_and_run(10)

        self.assertEqual(2, self.machine.ball_devices["bd_trough_front"].balls)
        self.assertEqual(2, self.machine.ball_devices["bd_trough_back"].balls)
        self.assertEqual(1, self.machine.playfields["playfield_front"].balls)
        self.assertEqual(1, self.machine.playfields["playfield_back"].balls)

        self.assertEqual("idle", self.machine.ball_devices["bd_trough_front"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_trough_back"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_launcher_front"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_launcher_back"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_feeder_front"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_feeder_back"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_middle_front"]._state)
        self.assertEqual("idle", self.machine.ball_devices["bd_middle_back"]._state)

    def testUnexpectedBallWithTransfer(self):
        # devices captures from one pf and ejects to another
        # tests target_on_unexpected_ball
        self.set_num_balls_known(4)
        self.machine.playfields["playfield_front"].balls = 2
        self.machine.playfields["playfield_front"].available_balls = 2
        self.machine.playfields["playfield_back"].balls = 2
        self.machine.playfields["playfield_back"].available_balls = 2

        self.hit_switch_and_run("s_middle_back1", 15)

        self.assertEqual(1, self.machine.playfields["playfield_front"].balls)
        self.assertEqual(1, self.machine.playfields["playfield_front"].available_balls)
        self.assertEqual(3, self.machine.playfields["playfield_back"].balls)
        self.assertEqual(3, self.machine.playfields["playfield_back"].available_balls)

    def testUnexpectedBallWithRouting(self):
        # device captures and ejects to same pf but ball has to routed through trough
        self.set_num_balls_known(4)
        self.machine.playfields["playfield_front"].balls = 2
        self.machine.playfields["playfield_front"].available_balls = 2
        self.machine.playfields["playfield_back"].balls = 2
        self.machine.playfields["playfield_back"].available_balls = 2

        self.hit_switch_and_run("s_launcher_lane_front", 1)

        # ball captured
        self.assertEqual(1, self.machine.playfields["playfield_front"].balls)
        self.assertEqual(2, self.machine.playfields["playfield_back"].balls)

        self.advance_time_and_run(20)
        self.assertEqual(2, self.machine.playfields["playfield_front"].balls)
        self.assertEqual(2, self.machine.playfields["playfield_back"].balls)

    def testPhantomballsAndGameStart(self):
        self.mock_event("playfield_jump")
        self._prepare_troughs()
        self.machine.playfields["playfield_front"].add_ball()
        self.machine.playfields["playfield_back"].add_ball()
        self.advance_time_and_run(10)

        # test pass from pf1 to pf2 but switch triggers twice
        pf1 = self.machine.ball_devices['playfield_front']
        pf2 = self.machine.ball_devices['playfield_back']

        self.assertEqual(1, pf1.balls)
        self.assertEqual(1, pf1.available_balls)
        self.assertEqual(1, pf2.balls)
        self.assertEqual(1, pf2.available_balls)

        # game should not start. there are balls on pf
        self.assertFalse(self.machine.ball_controller.request_to_start_game())

        self.machine.switch_controller.process_switch("s_transfer_front_back", 1)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("s_transfer_front_back", 0)
        self.advance_time_and_run(2)

        self.assertEqual(0, pf1.balls)
        self.assertEqual(0, pf1.available_balls)
        self.assertEqual(2, pf2.balls)
        self.assertEqual(2, pf2.available_balls)

        # ball drains in front
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_trough_front"])
        self.advance_time_and_run()

        self.assertEqual(0, pf1.balls)
        self.assertEqual(0, pf1.available_balls)
        self.assertEqual(1, pf2.balls)
        self.assertEqual(1, pf2.available_balls)

        # from the machine perspective one ball jumped from back to front
        self.assertEventCalledWith("playfield_jump", source=pf2, target=pf1)
        self.assertEventCalled("playfield_jump", 1)
        self.mock_event("playfield_jump")

        self.assertFalse(self.machine.ball_controller.request_to_start_game())

        # second ball also drains
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_trough_back"])
        self.advance_time_and_run()

        self.assertEqual(0, pf1.balls)
        self.assertEqual(0, pf1.available_balls)
        self.assertEqual(0, pf2.balls)
        self.assertEqual(0, pf2.available_balls)

        # game should start
        self.assertTrue(self.machine.ball_controller.request_to_start_game())
        self.assertEventNotCalled("playfield_jump")

    def testPlayfieldJump(self):
        self.mock_event("playfield_jump")
        self._prepare_troughs()
        self.machine.playfields["playfield_front"].add_ball()
        self.advance_time_and_run(10)

        # only front has a ball
        pf1 = self.machine.ball_devices['playfield_front']
        pf2 = self.machine.ball_devices['playfield_back']

        self.assertEqual(1, pf1.balls)
        self.assertEqual(1, pf1.available_balls)
        self.assertEqual(0, pf2.balls)
        self.assertEqual(0, pf2.available_balls)

        # balls jumps to back pf and falls into a device
        self.hit_switch_and_run("s_feeder_back", 5)

        self.assertEqual(0, pf1.balls)
        self.assertEqual(0, pf1.available_balls)
        self.assertEqual(1, pf2.balls)
        self.assertEqual(1, pf2.available_balls)

        self.assertEventCalledWith("playfield_jump", source=pf1, target=pf2)



"""Block Event Config Player."""
from typing import List

from mpf.core.config_player import ConfigPlayer


class BlockEventPlayer(ConfigPlayer):

    """Posts events based on config."""

    config_file_section = 'blocking'

    __slots__ = []  # type: List[str]

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Block event."""
        min_priority = kwargs.get("_min_priority", {"all": 0})

        for facility, min_priority_facility in settings.items():
            if min_priority_facility is True:
                min_priority_facility = priority
            elif min_priority_facility > priority:
                raise AssertionError("Cannot block above own priority. {} {}".format(facility, min_priority_facility))
            if facility not in min_priority or min_priority[facility] < min_priority_facility:
                min_priority[facility] = min_priority_facility

        return {"_min_priority": min_priority}

    def validate_config_entry(self, settings: dict, name: str) -> dict:
        """Validate one entry of this player."""
        return settings

    def get_express_config(self, value):
        """Parse short config."""
        raise AssertionError()



"""Contains the Tilt mode code."""
from typing import List, Any, Optional

from mpf.core.events import EventHandlerKey
from mpf.core.events import QueuedEvent
from mpf.core.mode import Mode


class Tilt(Mode):

    """A mode which handles a tilt in a pinball machine.

    Note that this mode is always running (even during attract mode) since the
    machine needs to watch for slam tilts at all times.
    """

    __slots__ = ["_balls_to_collect", "_last_warning", "ball_ending_tilted_queue", "tilt_event_handlers",
                 "last_tilt_warning_switch", "tilt_config", "_settle_time", "_warnings_to_tilt", "_multiple_hit_window"]

    def __init__(self, *args, **kwargs) -> None:
        """Create mode."""
        self._balls_to_collect = None           # type: Optional[int]
        self._last_warning = None               # type: Optional[int]
        self.ball_ending_tilted_queue = None    # type: Optional[QueuedEvent]
        self.tilt_event_handlers = []           # type: List[EventHandlerKey]
        self.last_tilt_warning_switch = None    # type: Optional[int]
        self.tilt_config = None                 # type: Any
        self._settle_time = None
        self._warnings_to_tilt = None
        self._multiple_hit_window = None
        super().__init__(*args, **kwargs)

    def mode_init(self):
        """Initialise mode."""
        self._balls_to_collect = 0
        self._last_warning = None
        self.ball_ending_tilted_queue = None
        self.tilt_event_handlers = []
        self.last_tilt_warning_switch = 0

        self.tilt_config = self.machine.config_validator.validate_config(
            config_spec='tilt',
            source=self.config.get('tilt', {}),
            section_name='tilt')

    def mode_start(self, **kwargs):
        """Start mode."""
        self._register_switch_handlers()

        for event in self.tilt_config['reset_warnings_events']:
            self.add_mode_event_handler(event, self.reset_warnings)

        for event in self.tilt_config['tilt_events']:
            self.add_mode_event_handler(event, self.tilt)

        for event in self.tilt_config['tilt_warning_events']:
            self.add_mode_event_handler(event, self.tilt_warning)

        for event in self.tilt_config['tilt_slam_tilt_events']:
            self.add_mode_event_handler(event, self.slam_tilt)

        self._settle_time = self.tilt_config['settle_time'].evaluate([])
        self._warnings_to_tilt = self.tilt_config['warnings_to_tilt'].evaluate([])
        self._multiple_hit_window = self.tilt_config['multiple_hit_window'].evaluate([])

    def mode_stop(self, **kwargs):
        """Stop mode."""
        self._remove_switch_handlers()

    def _register_switch_handlers(self):
        for switch in self.machine.switches.items_tagged(
                self.tilt_config['tilt_warning_switch_tag']):
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch,
                callback=self._tilt_warning_switch_handler)

        for switch in self.machine.switches.items_tagged(
                self.tilt_config['tilt_switch_tag']):
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch,
                callback=self.tilt)

        for switch in self.machine.switches.items_tagged(
                self.tilt_config['slam_tilt_switch_tag']):
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch,
                callback=self.slam_tilt)

    def _remove_switch_handlers(self):
        for switch in self.machine.switches.items_tagged(
                self.tilt_config['tilt_warning_switch_tag']):
            self.machine.switch_controller.remove_switch_handler(
                switch_name=switch.name,
                callback=self._tilt_warning_switch_handler)
        for switch in self.machine.switches.items_tagged(
                self.tilt_config['tilt_switch_tag']):
            self.machine.switch_controller.remove_switch_handler(
                switch_name=switch.name,
                callback=self.tilt)
        for switch in self.machine.switches.items_tagged(
                self.tilt_config['slam_tilt_switch_tag']):
            self.machine.switch_controller.remove_switch_handler(
                switch_name=switch.name,
                callback=self.slam_tilt)

    # ignore false positives about self.player
    # pylint: disable-msg=unsubscriptable-object,unsupported-assignment-operation
    def tilt_warning(self, **kwargs):
        """Process a tilt warning.

        If the number of warnings is the number to cause a tilt, a tilt will be
        processed.
        """
        del kwargs
        self.last_tilt_warning_switch = self.machine.clock.get_time()

        if (not self.machine.game or not self.machine.game.player or
                self.machine.game.ending or self.machine.game.tilted):
            return

        self.info_log("Tilt Warning")

        self._last_warning = self.machine.clock.get_time()
        self.machine.game.player[self.tilt_config['tilt_warnings_player_var']] += 1

        warnings = self.machine.game.player[self.tilt_config['tilt_warnings_player_var']]

        warnings_to_tilt = self._warnings_to_tilt
        if warnings >= warnings_to_tilt:
            self.tilt()
        else:
            self.machine.events.post(
                'tilt_warning',
                warnings=warnings,
                warnings_remaining=warnings_to_tilt - warnings)
            '''event: tilt_warning
            desc: A tilt warning just happened.
            args:
            warnings: The total number of warnings so far.
            warnings_remaining: The remaining number of warnings until a tilt.
            '''
            self.machine.events.post('tilt_warning_{}'.format(warnings))
            '''event: tilt_warning_(number)
            desc: A tilt warning just happened. The number of this tilt
            warning is in the event name in the (number).'''

    def reset_warnings(self, **kwargs):
        """Reset the tilt warnings for the current player."""
        del kwargs
        if not self.machine.game or not self.machine.game.player or self.machine.game.ending:
            return

        try:
            self.machine.game.player[self.tilt_config['tilt_warnings_player_var']] = 0
        except AttributeError:
            pass

    def tilt(self, **kwargs):
        """Cause the ball to tilt.

        This will post an event called *tilt*, set the game mode's ``tilted``
        attribute to *True*, disable the flippers and autofire devices, end the
        current ball, and wait for all the balls to drain.
        """
        del kwargs
        if not self.machine.game or self.machine.game.tilted or self.machine.game.ending:
            return

        self.machine.game.tilted = True

        self._balls_to_collect = 0
        for device in self.machine.ball_devices.values():
            if device.is_playfield():
                self._balls_to_collect += device.available_balls

        self.info_log("Processing Tilt. Balls to collect: %s",
                      self._balls_to_collect)

        self.machine.events.post('tilt')
        '''event: tilt
        desc: The player has tilted.'''

        self.tilt_event_handlers.append(
            self.machine.events.add_handler('player_turn_ending', self._ball_ending_tilted))

        for device in self.machine.ball_devices.values():
            if 'drain' in device.tags:
                self.tilt_event_handlers.append(
                    self.machine.events.add_handler(
                        'balldevice_{}_ball_enter'.format(device.name),
                        self._tilted_ball_drain))

        self.machine.game.end_ball()

    def _tilted_ball_drain(self, new_balls, unclaimed_balls, device, **kwargs):
        del new_balls
        del device
        del kwargs

        self._balls_to_collect -= unclaimed_balls

        self.debug_log("Tilted ball drain. Balls to collect: %s",
                       self._balls_to_collect)

        if self._balls_to_collect <= 0:
            self._tilt_done()

    def _tilt_switch_handler(self):
        self.tilt()

    def _tilt_warning_switch_handler(self):
        if (not self._last_warning or
                (self._last_warning + (self._multiple_hit_window * 0.001) <=
                 self.machine.clock.get_time())):

            self.tilt_warning()

    def _ball_ending_tilted(self, queue, **kwargs):
        del kwargs
        self.ball_ending_tilted_queue = queue
        queue.wait()

        if not self._balls_to_collect:
            self._tilt_done()

    def _tilt_done(self):
        if self.tilt_settle_ms_remaining:
            self.delay.reset(ms=self.tilt_settle_ms_remaining,
                             callback=self._tilt_done,
                             name='tilt')

        else:
            self.machine.game.tilted = False

            self.machine.events.post('tilt_clear')
            '''event: tilt_clear
            desc: Posted after a tilt, when the settling time has passed after
            the last tilt switch hit. This is used to hold the next ball
            start until the plumb bob has settled to prevent tilt throughs.
            '''

            if self.ball_ending_tilted_queue:
                self.ball_ending_tilted_queue.clear()
                self.ball_ending_tilted_queue = None

            self.machine.events.remove_handlers_by_keys(
                self.tilt_event_handlers)
            self.tilt_event_handlers = list()

    @property
    def tilt_settle_ms_remaining(self):
        """Return the amount of milliseconds remaining until the tilt settle time has cleared.

        Returns an integer of the number of ms remaining until tilt settled is cleared.
        """
        if not self.last_tilt_warning_switch:
            return 0

        delta = (self._settle_time -
                 (self.machine.clock.get_time() -
                  self.last_tilt_warning_switch) * 1000)
        if delta > 0:
            return delta

        return 0

    @property
    def tilt_warnings_remaining(self):
        """Return remaining warnings."""
        if not self.machine.game or not self.machine.game.player:
            return 0
        return self._warnings_to_tilt - self.machine.game.player[self.tilt_config['tilt_warnings_player_var']]

    def slam_tilt(self, **kwargs):
        """Process a slam tilt.

        This method posts the *slam_tilt* event and (if a game is active) sets
        the game mode's ``slam_tilted`` attribute to *True*.
        """
        del kwargs
        self.machine.events.post('slam_tilt')
        '''event: slam_tilt
        desc: A slam tilt has just occurred.'''
        if not self.machine.game:
            return

        self.machine.game.slam_tilted = True
        self.tilt()



from unittest.mock import MagicMock

from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestModes(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'test_modes.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/mode_tests/'

    def test_loading_modes(self):
        # Tests multi-case mode listing in config and checking here

        # assert that modes did load
        self.assertIn('mode1', self.machine.modes)
        self.assertIn('mode2', self.machine.modes)
        self.assertIn('mode3', self.machine.modes)
        self.assertIn('mode4', self.machine.modes)
        self.assertNotIn('mode8', self.machine.modes)

    def test_mode_start_stop(self):
        # Setup mocked event handlers for mode start/stop sequence
        self.mode1_will_start_event_handler = MagicMock()
        self.mode1_starting_event_handler = MagicMock()
        self.mode1_started_event_handler = MagicMock()
        self.mode1_will_stop_event_handler = MagicMock()
        self.mode1_stopping_event_handler = MagicMock()
        self.mode1_stopped_event_handler = MagicMock()

        self.machine.events.add_handler('mode_mode1_will_start', self.mode1_will_start_event_handler)
        self.machine.events.add_handler('mode_mode1_starting', self.mode1_starting_event_handler)
        self.machine.events.add_handler('mode_mode1_started', self.mode1_started_event_handler)
        self.machine.events.add_handler('mode_mode1_will_stop', self.mode1_will_stop_event_handler)
        self.machine.events.add_handler('mode_mode1_stopping', self.mode1_stopping_event_handler)
        self.machine.events.add_handler('mode_mode1_stopped', self.mode1_stopped_event_handler)

        # start mode 1. It should only start once
        self.machine.events.post('start_mode1')
        self.machine.events.post('start_mode1')
        self.advance_time_and_run()
        self.assertTrue(self.machine.mode_controller.is_active('mode1'))
        self.assertTrue(self.machine.modes["mode1"].active)
        self.assertIn(self.machine.modes["mode1"],
                      self.machine.mode_controller.active_modes)
        self.assertFalse(self.machine.mode_controller.is_active('mode2'))
        self.assertFalse(self.machine.modes["mode2"].active)
        self.assertIn(self.machine.modes["mode1"],
                      self.machine.mode_controller.active_modes)

        # test config via include
        self.assertEqual(123, self.machine.modes["mode1"].config['mode_settings']['test'])

        # start a mode that's already started and make sure it doesn't explode
        self.machine.modes["mode1"].start()

        # make sure event handler were called for mode start process
        self.assertEqual(1, self.mode1_will_start_event_handler.call_count)
        self.assertEqual(1, self.mode1_starting_event_handler.call_count)
        self.assertEqual(1, self.mode1_started_event_handler.call_count)

        # stop mode 1
        self.machine.events.post('stop_mode1')
        self.advance_time_and_run()
        self.assertFalse(self.machine.mode_controller.is_active('mode1'))
        self.assertFalse(self.machine.modes["mode1"].active)

        # make sure event handler were called for mode stop process
        self.assertEqual(1, self.mode1_will_stop_event_handler.call_count)
        self.assertEqual(1, self.mode1_stopping_event_handler.call_count)
        self.assertEqual(1, self.mode1_stopped_event_handler.call_count)

    def test_custom_mode_code(self):
        self.assertTrue(self.machine.modes["mode3"].custom_code)

    def test_mode_priorities(self):

        # test the start priorities TODO there's got to be a better way?
        found_it = False

        for handler in self.machine.events.registered_handlers['start_mode1']:

            if handler.callback == self.machine.modes["mode1"].start:
                found_it = True
                self.assertEqual(handler.priority, 201)

        self.assertTrue(found_it)

        # default priorities
        self.machine.modes["mode1"].start()
        self.advance_time_and_run()
        self.assertEqual(self.machine.modes["mode1"].priority, 200)
        self.assertEqual(self.machine.modes["attract"].priority, 10)
        self.assertEqual(self.machine.modes["mode1"].config['mode_settings']['this'], True)

        # test the stop priorities
        found_it = False
        for handler in self.machine.events.registered_handlers['stop_mode1']:
            if handler.callback == self.machine.modes["mode1"].stop:
                found_it = True
                # 201 because stop priorities are always +1 over mode priority
                self.assertEqual(handler.priority, 201)
        self.assertTrue(found_it)

        # pass a priority at start
        self.machine.modes["mode1"].stop()
        self.advance_time_and_run()
        self.machine.modes["mode1"].start(mode_priority=500)
        self.advance_time_and_run()
        self.assertEqual(self.machine.modes["mode1"].priority, 500)
        self.advance_time_and_run()

        # test the order of the active modes list
        self.assertEqual(self.machine.modes["mode1"],
                         self.machine.mode_controller.active_modes[0])
        self.assertEqual(self.machine.modes["attract"],
                         self.machine.mode_controller.active_modes[1])

    def test_mode_start_with_callback(self):
        self.mode_start_callback = MagicMock()

        self.machine.modes["mode1"].start(callback=self.mode_start_callback)
        self.advance_time_and_run()
        self.mode_start_callback.assert_called_once_with()

    def test_use_wait_queue(self):
        self.callback = MagicMock()
        self.machine.events.post_queue('start_mode4',
                                       callback=self.callback)
        self.advance_time_and_run()

        # make sure the mode started
        self.assertTrue(self.machine.modes["mode4"].active)
        self.advance_time_and_run()

        # callback should not be called since mode is set to use wait queue
        self.callback.assert_not_called()

        self.machine.modes["mode4"].stop()
        self.advance_time_and_run()

        # once mode stops, verify that callback was called
        self.callback.assert_called_once_with()

    def test_ball_end(self):
        # mode1 is set to keep running on ball end
        # mode2 should stop and restart on ball end
        # mode3 should stop and stay stopped
        self.assertFalse(self.machine.modes["mode1"].config['mode']['stop_on_ball_end'])
        self.assertTrue(self.machine.modes["mode2"].config['mode']['stop_on_ball_end'])
        self.assertTrue(self.machine.modes["mode3"].config['mode']['stop_on_ball_end'])

        # start a game
        self.machine.playfield.add_ball = MagicMock()
        self.machine.events.post('game_start')
        self.advance_time_and_run()
        self.machine.game.balls_in_play = 1
        self.assertTrue(self.machine.game)

        # start some modes
        self.machine.modes["mode1"].start()
        self.machine.modes["mode2"].start()
        self.machine.modes["mode3"].start()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["mode1"].active)
        self.assertTrue(self.machine.modes["mode2"].active)
        self.assertTrue(self.machine.modes["mode3"].active)

        # change the config of modes 1 and 2 so we can verify whether they
        # actually stopped and restarted since the configs are used to start
        # the modes
        self.machine.modes["mode1"].config['mode']['priority'] = 999
        self.machine.modes["mode2"].config['mode']['priority'] = 999

        # end the ball
        self.machine.game.end_ball()
        self.advance_time_and_run()

        self.assertTrue(self.machine.modes["mode1"].active)
        self.assertTrue(self.machine.modes["mode2"].active)
        self.assertFalse(self.machine.modes["mode3"].active)

        # since mode1 should have stayed running, its config should not change
        self.assertNotEqual(self.machine.modes["mode1"].priority, 999)
        # mode2 should use the new priority since it should have restarted
        self.assertEqual(self.machine.modes["mode2"].priority, 999)

        # end ball 2
        self.machine.game.end_ball()
        self.advance_time_and_run()

        # end ball 3 and end the game
        self.machine.game.end_ball()
        self.advance_time_and_run()

        self.assertTrue(self.machine.modes["attract"].active)
        self.assertFalse(self.machine.modes["game"].active)
        self.assertTrue(self.machine.modes["mode1"].active)
        self.assertFalse(self.machine.modes["mode2"].active)
        self.assertFalse(self.machine.modes["mode3"].active)


class TestModesInGame(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'test_modes_in_game.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/mode_tests/'

    def test_restart_on_next_ball(self):
        """Test restart_on_next_ball."""
        self.mock_event("mode_mode_restart_on_next_ball_will_start")
        self.assertModeNotRunning("mode_restart_on_next_ball")
        self.start_game()

        self.assertModeNotRunning("mode_restart_on_next_ball")
        self.drain_all_balls()

        # mode shoud not be started
        self.assertModeNotRunning("mode_restart_on_next_ball")

        # start it
        self.post_event("start_mode_restart_on_next_ball")
        self.advance_time_and_run()

        # it should run
        self.assertModeRunning("mode_restart_on_next_ball")
        self.assertEventCalled("mode_mode_restart_on_next_ball_will_start", 1)

        # check that mode is restarted on next ball
        self.mock_event("mode_mode_restart_on_next_ball_will_start")
        self.drain_all_balls()
        self.assertModeRunning("mode_restart_on_next_ball")
        self.assertEventCalled("mode_mode_restart_on_next_ball_will_start", 1)



import sys

from mpf.tests.MpfBcpTestCase import MpfBcpTestCase
from unittest.mock import create_autospec, patch, MagicMock


class TestServiceCli(MpfBcpTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/service_mode/'

    def _table(self, data):
        result = ""
        for row in data:
            for col in row:
                result += str(col) + ";"
            result += "\n"
        obj = MagicMock()
        obj.table = result[:-1]
        return obj

    def setUp(self):
        modules = {
            'terminaltables': MagicMock(),
        }
        modules["terminaltables"].AsciiTable = self._table

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        super().setUp()
        self.mock_stdin = create_autospec(sys.stdin)
        self.mock_stdout = create_autospec(sys.stdout)
        # we are connected as anonymous client
        self._bcp_client.name = None

    def tearDown(self) -> None:
        super().tearDown()
        self.module_patcher.stop()

    def _last_write(self, n=None):
        """:return: last `n` output lines"""
        if n is None:
            return self.mock_stdout.write.call_args[0][0]
        return "".join(map(lambda c: c[0][0], self.mock_stdout.write.call_args_list[-n:]))

    def test_cli(self):
        self.maxDiff = None
        with patch("mpf.commands.service.AsciiTable"):
            from mpf.commands.service import ServiceCli
            cli = ServiceCli(self._bcp_external_client, self.loop, self.mock_stdin, self.mock_stdout)

        cli.onecmd("light_color l_light5 red")
        self.assertEqual("Success\n", self._last_write())

        cli.onecmd("light_color l_light1 invalid_color")
        self.assertEqual("Error: Invalid RGB string: invalid_color\n", self._last_write())

        cli.onecmd("light_color l_light1 white")
        self.assertEqual("Success\n", self._last_write())

        cli.onecmd("list_lights")
        expected = """Board;Number;Name;Color;
Virtual;['led-1-b', 'led-1-g', 'led-1-r'];l_light1;(255, 255, 255);
Virtual;['led-5-b', 'led-5-g', 'led-5-r'];l_light5;(255, 0, 0);
"""
        self.assertEqual(expected, self._last_write())

        cli.onecmd("light_off l_light5")
        self.assertEqual("Success\n", self._last_write())

        self.assertLightColor("l_light5", "black")

        cli.onecmd("light_color l_light5")
        self.assertEqual("Success\n", self._last_write())

        self.assertLightColor("l_light5", "white")

        cli.onecmd("light_off l_light5")
        self.assertEqual("Success\n", self._last_write())

        self.assertLightColor("l_light5", "black")

        self.hit_switch_and_run("s_door_open", 0)

        cli.onecmd("list_switches")
        expected = """Board;Number;Name;State;
Virtual;1;s_door_open;closed;
Virtual;17;s_service_enter;open;
Virtual;18;s_service_esc;open;
Virtual;19;s_service_up;open;
Virtual;20;s_service_down;open;
"""
        self.assertEqual(expected, self._last_write())

        cli.onecmd("list_coils")
        expected = """Board;Number;Name;
Virtual;1;c_test;
Virtual;2;c_test2;
Virtual;3;c_test5;
Virtual;10;c_test6;
Virtual;100;c_test4;
Virtual;1000;c_test3;
"""
        self.assertEqual(expected, self._last_write())

        cli.onecmd("list_shows")
        expected = """Name;Token;
flash;['led', 'leds', 'light', 'lights'];
flash_color;['color', 'led', 'leds', 'light', 'lights'];
led_color;['color', 'led', 'leds', 'light', 'lights'];
off;['led', 'leds', 'light', 'lights'];
on;['led', 'leds', 'light', 'lights'];
"""
        self.assertEqual(expected, self._last_write())

        self.assertLightColor("l_light5", "black")

        cli.onecmd("show_play on led:l_light5")
        self.assertLightColor("l_light5", "white")

        cli.onecmd("show_stop on")
        self.assertLightColor("l_light5", "black")

        cli.onecmd("show_play led_color led:l_light5 color:red")
        self.assertLightColor("l_light5", "red")

        self.assertEqual("disabled", self.machine.coils["c_test"].hw_driver.state)

        cli.onecmd("coil_pulse c_test")
        self.assertEqual("Success\n", self._last_write())
        self.assertEqual("pulsed_10", self.machine.coils["c_test"].hw_driver.state)

        cli.onecmd("coil_enable c_test")
        self.assertEqual("Error: Cannot enable driver with hold_power 0.0\n", self._last_write())

        cli.onecmd("coil_enable c_test6")
        self.assertEqual("Success\n", self._last_write())
        self.assertEqual("enabled", self.machine.coils["c_test6"].hw_driver.state)

        cli.onecmd("coil_disable c_test6")
        self.assertEqual("Success\n", self._last_write())
        self.assertEqual("disabled", self.machine.coils["c_test6"].hw_driver.state)

        self.assertLightColor("l_light1", "white")
        self.assertLightColor("l_light5", "red")

        cli.onecmd("quit")
        self.advance_time_and_run()

        self.assertLightColor("l_light1", "black")
        self.assertLightColor("l_light5", "black")



"""Platform to control the hardware of a Raspberry Pi."""
import asyncio

from typing import Optional

from mpf.platforms.interfaces.i2c_platform_interface import I2cPlatformInterface

from mpf.core.delays import DelayManager
from mpf.core.utility_functions import Util

from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface
from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings
from mpf.core.platform import SwitchPlatform, DriverPlatform, ServoPlatform, SwitchSettings, \
    DriverSettings, DriverConfig, SwitchConfig, I2cPlatform, RepulseSettings

# apiogpio is not a requirement for MPF so we fail with a nice error when loading
try:
    import apigpio
except ImportError:
    apigpio = None


class RpiSwitch(SwitchPlatformInterface):

    """A switch on a RPI."""

    def get_board_name(self):
        """Return name."""
        return "Raspberry Pi"


class RpiDriver(DriverPlatformInterface):

    """An output on a Rasoberry Pi."""

    def __init__(self, number, config, platform):
        """Initialise output."""
        super().__init__(config, number)
        self.platform = platform            # type: RaspberryPiHardwarePlatform
        self.gpio = int(self.number)
        self.delay = DelayManager(self.platform.machine)

    def get_board_name(self):
        """Return name."""
        return "Raspberry Pi"

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse output."""
        self.enable(pulse_settings, None)

    def enable(self, pulse_settings: PulseSettings, hold_settings: Optional[HoldSettings]):
        """Enable output."""
        self.platform.send_command(self.platform.pi.write(self.gpio, 1))
        if hold_settings and hold_settings.power == 1:
            # do nothing. just keep driver enabled
            pass
        elif hold_settings and hold_settings.power > 0:
            # schedule pwm
            self.delay.add(pulse_settings.duration, self._pwm, hold_power=hold_settings.power)
        else:
            # no hold. disable after pulse
            self.delay.add(pulse_settings.duration, self.disable)

    def _pwm(self, hold_power):
        """Set up pwm."""
        self.platform.send_command(self.platform.pi.set_PWM_dutycycle(self.gpio, hold_power * 255))

    def disable(self):
        """Disable output."""
        self.platform.send_command(self.platform.pi.write(self.gpio, 0))
        # clear all delays
        self.delay.clear()


class RpiServo(ServoPlatformInterface):

    """A servo connected to a RPI."""

    def __init__(self, number, platform):
        """Initialise servo."""
        self.gpio = int(number)
        self.platform = platform    # type: RaspberryPiHardwarePlatform

    def go_to_position(self, position):
        """Move servo to position."""
        # can be between 1000 and 2000us
        position_translated = 1000 + position * 1000
        self.platform.send_command(self.platform.pi.set_servo_pulsewidth(self.gpio, position_translated))

    def set_speed_limit(self, speed_limit):
        """Not implemented."""

    def set_acceleration_limit(self, acceleration_limit):
        """Not implemented."""


class RpiI2cDevice(I2cPlatformInterface):

    """A I2c device on a Rpi."""

    def __init__(self, number: str, loop, platform) -> None:
        """Initialise i2c device on rpi."""
        super().__init__(number)
        self.loop = loop
        self.pi = platform.pi
        self.platform = platform
        self.number = number
        self.handle = None

    async def open(self):
        """Open I2c port."""
        self.handle = await self._get_i2c_handle(self.number)

    @staticmethod
    def _get_i2c_bus_address(address):
        """Split and return bus + address."""
        if isinstance(address, int):
            return 0, address
        bus, address = address.split("-")
        return int(bus), int(address)

    async def _get_i2c_handle(self, address):
        """Get or open handle for i2c device via pigpio."""
        bus_address, device_address = self._get_i2c_bus_address(address)
        handle = await self.pi.i2c_open(bus_address, device_address)
        return handle

    def i2c_write8(self, register, value):
        """Write to i2c via pigpio."""
        self.platform.send_command(self._i2c_write8_async(register, value))

    async def _i2c_write8_async(self, register, value):
        await self.pi.i2c_write_byte_data(self.handle, register, value)

    async def i2c_read8(self, register):
        """Read from i2c via pigpio."""
        return await self.pi.i2c_read_byte_data(self.handle, register)

    async def i2c_read_block(self, register, count):
        """Read block via I2C."""
        return await self.pi.i2c_read_i2c_block_data(self.handle, register, count)


class RaspberryPiHardwarePlatform(SwitchPlatform, DriverPlatform, ServoPlatform, I2cPlatform):

    """Control the hardware of a Raspberry Pi.

    Works locally and remotely via network.
    """

    def __init__(self, machine):
        """Initialise Raspberry Pi platform."""
        super().__init__(machine)

        if not apigpio:
            raise AssertionError("To use the Raspberry Pi platform you need to install the apigpio extension. "
                                 "Run: pip3 install apigpio-mpf.")

        self.pi = None          # type: Optional[apigpio.Pi]
        # load config
        self.config = self.machine.config_validator.validate_config("raspberry_pi", self.machine.config['raspberry_pi'])
        self._switches = None   # type: Optional[int]

        self._cmd_queue = None  # type: Optional[asyncio.Queue]
        self._cmd_task = None   # type: Optional[asyncio.Task]
        self._configure_device_logging_and_debug("Raspberry Pi", self.config)

    async def initialize(self):
        """Initialise platform."""
        # create pi object and connect
        self.pi = apigpio.Pi(self.machine.clock.loop)
        await self.pi.connect((self.config['ip'], self.config['port']))

        self._switches = await self.pi.read_bank_1()

        self._cmd_queue = asyncio.Queue()
        self._cmd_task = self.machine.clock.loop.create_task(self._run())
        self._cmd_task.add_done_callback(Util.raise_exceptions)

    def send_command(self, cmd):
        """Add a command to the command queue."""
        self._cmd_queue.put_nowait(cmd)

    async def _run(self):
        """Handle the command queue."""
        while True:
            # get next command
            cmd = await self._cmd_queue.get()
            # run command
            await cmd

    def stop(self):
        """Stop platform."""
        if self._cmd_task:
            self._cmd_task.cancel()
            self._cmd_task = None

        if self.pi:
            self.machine.clock.loop.run_until_complete(self.pi.stop())
            self.pi = None

    async def configure_servo(self, number: str) -> ServoPlatformInterface:
        """Configure a servo."""
        return RpiServo(number, self)

    async def get_hw_switch_states(self):
        """Return current switch states."""
        hw_states = dict()
        curr_bit = 1
        for index in range(32):
            hw_states[str(index)] = bool(curr_bit & self._switches)
            curr_bit <<= 1
        return hw_states

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> "SwitchPlatformInterface":
        """Configure a switch with pull up."""
        # set input
        self.send_command(self.pi.set_mode(int(number), apigpio.INPUT))
        # configure pull up
        self.send_command(self.pi.set_pull_up_down(int(number), apigpio.PUD_UP))

        # if config.debounce:
        #     # configure debounce to 2ms
        #     self.send_command(self.pi.set_glitch_filter(int(number), 2000))
        # else:
        #     # configure debounce to 100us
        #     self.send_command(self.pi.set_glitch_filter(int(number), 100))

        # add callback
        self.send_command(self.pi.add_callback(int(number), apigpio.EITHER_EDGE, self._switch_changed))

        return RpiSwitch(config, number)

    def _switch_changed(self, gpio, level, tick):
        """Process switch change."""
        del tick
        self.machine.switch_controller.process_switch_by_num(str(gpio), level, self)

    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Raise exception."""
        raise AssertionError("Not supported on the RPi currently. Write in the forum if you need it.")

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Raise exception."""
        raise AssertionError("Not supported on the RPi currently. Write in the forum if you need it.")

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                      eos_switch: SwitchSettings, coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Raise exception."""
        raise AssertionError("Not supported on the RPi currently. Write in the forum if you need it.")

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Raise exception."""
        raise AssertionError("Not supported on the RPi currently. Write in the forum if you need it.")

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Raise exception."""
        raise AssertionError("Not supported on the RPi currently. Write in the forum if you need it.")

    def clear_hw_rule(self, switch: SwitchSettings, coil: DriverSettings):
        """Raise exception."""
        raise AssertionError("Not supported on the RPi currently. Write in the forum if you need it.")

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict) -> "DriverPlatformInterface":
        """Configure an output on the Raspberry Pi."""
        # disable pull up/down
        self.send_command(self.pi.set_pull_up_down(int(number), apigpio.PUD_OFF))
        # set output
        self.send_command(self.pi.set_mode(int(number), apigpio.OUTPUT))

        return RpiDriver(number, config, self)

    async def configure_i2c(self, number: str):
        """Configure I2c device."""
        device = RpiI2cDevice(number, self.machine.clock.loop, self)
        await device.open()
        return device



"""Pololu TIC Device."""
import subprocess
import logging
import asyncio

from threading import Thread
import ruamel.yaml

from mpf.core.utility_functions import Util


class TicError(Exception):

    """A Pololu TIC Error."""


class PololuTiccmdWrapper:

    """A Pololu TIC Device."""

    def __init__(self, serial_number, machine, debug=True):
        """Return the current status of the TIC device.

        Args:
        ----
            serial_number (number): The serial number of the TIC to control
            machine (object): The machine object
            debug (boolean): Turn on debugging or not
        """
        self._debug = debug
        self.log = logging.getLogger('TIC Stepper')
        self._serial_number = serial_number
        self._machine = machine
        self.loop = None

        self._start_thread()

    def _start_thread(self):
        # Create a new loop
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.stop_future = asyncio.Future()
        # Assign the loop to another thread
        self.thread = Thread(target=self._run_loop)
        self.thread.start()

    def _run_loop(self):
        """Run the asyncio loop in this thread."""
        self.loop.run_until_complete(self.stop_future)
        self.loop.close()

    def stop_async(self):
        """Stop loop."""
        self.stop_future.set_result(True)

    def stop(self):
        """Stop loop and join thread."""
        self.loop.call_soon_threadsafe(self.stop_async)
        self._stop_thread()

    def _stop_thread(self):
        self.thread.join()

    def _ticcmd_future(self, *args):
        """Run ticcmd in another thread."""
        future = asyncio.wrap_future(
            asyncio.run_coroutine_threadsafe(self._ticcmd_remote(*args), self.loop))
        future.add_done_callback(Util.raise_exceptions)
        return future

    def _ticcmd(self, *args):
        """Run ticcmd in another thread and forget about the response."""
        self.loop.call_soon_threadsafe(self._run_subprocess_ticcmd, *args)

    async def _ticcmd_remote(self, *args):
        """Return a future with the result of ticcmd."""
        return self._run_subprocess_ticcmd(*args)

    def _run_subprocess_ticcmd(self, *args):
        """Run ticcmd.

        This will block the asyncio loop in the thread so only one command can run at a time.
        However, it will not block MPF because we will call this via run_coroutine_threadsafe from another thread.
        """
        args = list(args)
        args.append('-d')
        args.append(str(self._serial_number))
        try:
            output = subprocess.check_output(['ticcmd'] + args, stderr=subprocess.STDOUT)
            return output
        except subprocess.CalledProcessError as e:
            self.log.debug("Exception: %s", str(e.output))
            raise TicError(e.output)

    async def get_status(self):
        """Return the current status of the TIC device."""
        cmd_return = await self._ticcmd_future('-s', '--full')
        status = ruamel.yaml.safe_load(cmd_return)
        return status

    def halt_and_hold(self):
        """Stop the motor abruptly without respecting the deceleration limit."""
        self._ticcmd('--halt-and-hold')

    def halt_and_set_position(self, position):
        """Stop the motor abruptly without respecting the deceleration limit and sets the current position."""
        self._ticcmd('--halt-and-set-position', str(int(position)))

    def rotate_to_position(self, position):
        """Tells the TIC to move the stepper to the target position.

        Args:
        ----
            position (number): The desired position in microsteps
        """
        self._ticcmd('--position', str(int(position)))

    def rotate_by_velocity(self, velocity):
        """Tells the TIC to move the stepper continuously at the specified velocity.

        Args:
        ----
            velocity (number): The desired speed in microsteps per 10,000 s
        """
        self._ticcmd('--velocity', str(int(velocity)))

    def reset_command_timeout(self):
        """Tells the TIC to reset the internal command timeout."""
        self._ticcmd('--reset-command-timeout')

    def exit_safe_start(self):
        """Tells the TIC to exit the safe start mode."""
        self._ticcmd('--exit-safe-start')

    def set_step_mode(self, mode):
        """Set the Step Mode of the stepper.

        Args:
        ----
            mode (number): One of 1, 2, 4, 8, 16, 32, the number of microsteps per step
        """
        self._ticcmd('--step-mode', str(int(mode)))

    def set_max_speed(self, speed):
        """Set the max speed of the stepper.

        Args:
        ----
            speed (number): The maximum speed of the stepper in microsteps per 10,000s
        """
        self._ticcmd('--max-speed', str(int(speed)))

    def set_starting_speed(self, speed):
        """Set the starting speed of the stepper.

        Args:
        ----
            speed (number): The starting speed of the stepper in microsteps per 10,000s
        """
        self._ticcmd('--starting-speed', str(int(speed)))

    def set_max_acceleration(self, acceleration):
        """Set the max acceleration of the stepper.

        Args:
        ----
            acceleration (number): The maximum acceleration of the stepper in microsteps per 100 s^2
        """
        self._ticcmd('--max-accel', str(int(acceleration)))

    def set_max_deceleration(self, deceleration):
        """Set the max deceleration of the stepper.

        Args:
        ----
            deceleration (number): The maximum deceleration of the stepper in microsteps per 100 s^2
        """
        self._ticcmd('--max-decel', str(int(deceleration)))

    def set_current_limit(self, current):
        """Set the max current of the stepper driver.

        Args:
        ----
            current (number): The maximum current of the stepper in milliamps
        """
        self._ticcmd('--current', str(int(current)))

    def energize(self):
        """Energize the Stepper."""
        self._ticcmd('--energize')

    def go_home(self, forward):
        """Energize the Stepper."""
        if forward:
            direction = "fwd"
        else:
            direction = "rev"
        self._ticcmd('--home {}'.format(direction))



"""Contains the ShotGroup base class."""

from collections import deque

from mpf.core.device_monitor import DeviceMonitor

from mpf.core.events import event_handler
from mpf.core.mode import Mode
from mpf.core.mode_device import ModeDevice
from mpf.core.player import Player


@DeviceMonitor("common_state", "rotation_enabled")
class ShotGroup(ModeDevice):

    """Represents a group of shots in a pinball machine by grouping together multiple `Shot` class devices.

    This is used so you get get
    "group-level" functionality, like shot rotation, shot group completion,
    etc. This would be used for a group of rollover lanes, a bank of standups,
    etc.
    """

    config_section = 'shot_groups'
    collection = 'shot_groups'
    class_label = 'shot_group'

    __slots__ = ["rotation_enabled", "profile", "rotation_pattern"]

    def __init__(self, machine, name):
        """Initialise shot group."""
        super().__init__(machine, name)

        self.rotation_enabled = None
        self.profile = None
        self.rotation_pattern = None

    def add_control_events_in_mode(self, mode) -> None:
        """Remove enable here."""

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Add device in mode."""
        super().device_loaded_in_mode(mode, player)
        self._check_for_complete()
        self.profile = self.config['shots'][0].profile
        self.rotation_pattern = deque(self.profile.config['rotation_pattern'])
        self.rotation_enabled = not self.config['enable_rotation_events']
        for shot in self.config['shots']:
            self.machine.events.add_handler("{}_hit".format(shot.name), self._hit)

    def device_removed_from_mode(self, mode):
        """Disable device when mode stops."""
        super().device_removed_from_mode(mode)
        self.machine.events.remove_handler(self._hit)

    @property
    def common_state(self):
        """Return common state if all shots in this group are in the same state.

        Will return None otherwise.
        """
        state = self.config['shots'][0].state_name
        for shot in self.config['shots']:
            if state != shot.state_name:
                # shots do not have a common state
                return None

        return state

    def _check_for_complete(self):
        """Check if all shots in this group are in the same state."""
        state = self.common_state
        if not state:
            # shots do not have a common state
            return

        # if we reached this point we got a common state

        self.debug_log(
            "Shot group is complete with state: %s", state)

        self.machine.events.post('{}_complete'.format(self.name), state=state)
        '''event: (name)_complete
        desc: All the member shots in the shot group called (name)
        are in the same state.

        args:
          state: name of the common state of all shots.
        '''

        self.machine.events.post('{}_{}_complete'.format(self.name, state))
        '''event: (name)_(state)_complete
        desc: All the member shots in the shot group called (name)
        are in the same state named (state).
        '''

    @event_handler(2)
    def event_enable(self, **kwargs):
        """Handle enable control event."""
        del kwargs
        self.enable()

    def enable(self):
        """Enable all member shots."""
        for shot in self.config['shots']:
            shot.enable()

    @event_handler(3)
    def event_disable(self, **kwargs):
        """Handle disable control event."""
        del kwargs
        self.disable()

    def disable(self):
        """Disable all member shots."""
        for shot in self.config['shots']:
            shot.disable()

    @event_handler(1)
    def event_reset(self, **kwargs):
        """Handle reset control event."""
        del kwargs
        self.reset()

    def reset(self):
        """Reset all member shots."""
        for shot in self.config['shots']:
            shot.reset()

    @event_handler(4)
    def event_restart(self, **kwargs):
        """Handle restart control event."""
        del kwargs
        self.restart()

    def restart(self):
        """Restart all member shots."""
        for shot in self.config['shots']:
            shot.restart()

    def _hit(self, advancing, **kwargs):
        """One of the member shots in this shot group was hit.

        Args:
        ----
            kwarg: {
                profile: the current profile of the member shot that was hit
                state: the current state of the member shot that was hit
                advancing: boolean of whether the state is advancing
            }
        """
        if advancing:
            self._check_for_complete()

        self.machine.events.post(self.name + '_hit')
        '''event: (name)_hit
        desc: A member shots in the shot group called (name)
        has been hit.
        '''
        self.machine.events.post("{}_{}_hit".format(self.name, kwargs['state']))
        '''event: (name)_(state)_hit
        desc: A member shot with state (state) in the shot group (name)
        has been hit.
        '''

    @event_handler(9)
    def event_enable_rotation(self, **kwargs):
        """Handle enable_rotation control event."""
        del kwargs
        self.enable_rotation()

    def enable_rotation(self):
        """Enable shot rotation.

        If disabled, rotation events do not actually rotate the shots.
        """
        self.debug_log('Enabling rotation')
        self.rotation_enabled = True

    @event_handler(2)
    def event_disable_rotation(self, **kwargs):
        """Handle disable rotation control event."""
        del kwargs
        self.disable_rotation()

    def disable_rotation(self):
        """Disable shot rotation.

        If disabled, rotation events do not actually rotate the shots.
        """
        self.debug_log('Disabling rotation')
        self.rotation_enabled = False

    @event_handler(4)
    def event_rotate(self, direction=None, **kwargs):
        """Handle rotate control event."""
        del kwargs
        self.rotate(direction)

    def rotate(self, direction=None):
        """Rotate (or "shift") the state of all the shots in this group.

        This is used for things like lane change, where hitting the flipper
        button shifts all the states of the shots in the group to the left or
        right.

        This method actually transfers the current state of each shot profile
        to the left or the right, and the shot on the end rolls over to the
        taret on the other end.

        Args:
        ----
            direction: String that specifies whether the rotation direction is
                to the left or right. Values are 'right' or 'left'. Default of
                None will cause the shot group to rotate in the direction as
                specified by the rotation_pattern.

        Note that this shot group must, and rotation_events for this
        shot group, must both be enabled for the rotation events to work.
        """
        if not self.rotation_enabled:
            self.debug_log("Received rotation request. "
                           "Rotation Enabled: %s. Will NOT rotate",
                           self.rotation_enabled)

            return

        # shot_state_list is deque of tuples (state num, show step num)
        shot_state_list = deque()

        shots_to_rotate = []
        for shot in self.config['shots']:
            if shot.can_rotate:
                shots_to_rotate.append(shot)
                shot_state_list.append(shot.state)

        # figure out which direction we're going to rotate
        if not direction:
            direction = self.rotation_pattern[0]
            self.rotation_pattern.rotate(-1)
            self.debug_log("Since no direction was specified, pulling from"
                           " rotation pattern: '%s'", direction)

        # rotate that list
        if direction.lower() in ('right', 'r'):
            shot_state_list.rotate(1)
        else:
            shot_state_list.rotate(-1)

        # step through all our shots and update their states
        for i, shot in enumerate(shots_to_rotate):
            shot.jump(state=shot_state_list[i], force=True)

    @event_handler(8)
    def event_rotate_right(self, **kwargs):
        """Handle rotate right control event."""
        del kwargs
        self.rotate_right()

    def rotate_right(self):
        """Rotate the state of the shots to the right.

        This method is the same as calling rotate('right')
        """
        self.rotate(direction='right')

    @event_handler(7)
    def event_rotate_left(self, **kwargs):
        """Handle rotate left control event."""
        del kwargs
        self.rotate_left()

    def rotate_left(self):
        """Rotate the state of the shots to the left.

        This method is the same as calling rotate('left')
        """
        self.rotate(direction='left')



"""Contains the LogMixin class."""
import logging
from logging import Logger

from mpf.exceptions.config_file_error import ConfigFileError
from mpf._version import log_url

MYPY = False
if MYPY:   # pragma: no cover
    from typing import NoReturn, Optional  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class LogMixin:

    """Mixin class to add smart logging functionality to modules."""

    unit_test = False

    __slots__ = ["log", "_info_to_console", "_debug_to_console", "_debug", "_info_to_file", "_debug_to_file",
                 "_info", "_url_base"]

    def __init__(self) -> None:
        """Initialise Log Mixin."""
        self.log = None     # type: Optional[Logger]
        self._info_to_console = False
        self._debug_to_console = False
        self._info_to_file = False
        self._info = False
        self._debug_to_file = False
        self._debug = False
        self._url_base = None   # type: Optional[str]

        if MYPY:
            self.machine = self.machine     # type: MachineController # noqa

        logging.addLevelName(21, "INFO")
        logging.addLevelName(11, "DEBUG")
        logging.addLevelName(22, "INFO")
        logging.addLevelName(12, "DEBUG")

    def configure_logging(self, logger: str, console_level: str = 'basic',
                          file_level: str = 'basic', url_base=None):
        """Configure logging.

        Args:
        ----
            logger: The string name of the logger to use.
            console_level: The level of logging for the console. Valid options
                are "none", "basic", or "full".
            file_level: The level of logging for the console. Valid options
                are "none", "basic", or "full".
            url_base: Base URL for docs links in exceptions.
        """
        if isinstance(url_base, str):
            self._url_base = url_base
        else:
            self._url_base = logger
        self.log = logging.getLogger(logger)
        if hasattr(self, "machine") and self.machine and self.machine.options['production']:
            return

        try:
            if console_level.lower() == 'basic':
                self._info_to_console = True
                self._info = True
            elif console_level.lower() == 'full':
                self._debug_to_console = True
                self._debug = True
        except AttributeError:
            pass

        try:
            if file_level.lower() == 'basic':
                self._info_to_file = True
                self._info = True
            elif file_level.lower() == 'full':
                self._debug_to_file = True
                self._debug = True
        except AttributeError:
            pass

        # in unit tests always log info. debug will depend on the actual settings.
        if self.unit_test:
            self._info_to_console = True
            self._info = True

    def debug_log(self, msg: str, *args, context=None, error_no=None, **kwargs) -> None:
        """Log a message at the debug level.

        Note that whether this message shows up in the console or log file is
        controlled by the settings used with configure_logging().
        """
        if not self.log:
            self._logging_not_configured()

        if self._debug_to_console:
            level = 12
        elif self._debug_to_file:
            level = 11
        else:
            return

        if not self.log.isEnabledFor(level):
            return

        self.log.log(level, self.format_log_line(msg, context, error_no), *args, **kwargs)

    def info_log(self, msg: str, *args, context=None, error_no=None, **kwargs) -> None:
        """Log a message at the info level.

        Whether this message shows up in the console or log file is controlled
        by the settings used with configure_logging().
        """
        if not self.log:
            self._logging_not_configured()

        if self._info_to_console or self._debug_to_console:
            level = 22
        elif self._info_to_file or self._debug_to_file:
            level = 21
        else:
            return

        if not self.log.isEnabledFor(level):
            return

        self.log.log(level, self.format_log_line(msg, context, error_no), *args, **kwargs)

    def warning_log(self, msg: str, *args, context=None, error_no=None, **kwargs) -> None:
        """Log a message at the warning level.

        These messages will always be shown in the console and the log file.
        """
        if not self.log:
            self._logging_not_configured()

        if not self.log.isEnabledFor(30):
            return

        self.log.log(30, self.format_log_line(msg, context, error_no), *args, **kwargs)

    def error_log(self, msg: str, *args, context=None, error_no=None, **kwargs) -> None:
        """Log a message at the error level.

        These messages will always be shown in the console and the log file.
        """
        if not self.log:
            self._logging_not_configured()

        if not self.log.isEnabledFor(40):
            return

        self.log.log(40, self.format_log_line(msg, context, error_no), *args, **kwargs)

    def format_log_line(self, msg, context, error_no) -> str:
        """Return a formatted log line with log link and context."""
        if error_no:
            error_slug = "Log-{}-{}".format(self.log.name if self.log else "", error_no)
            error_url = log_url.format(error_slug)
        if error_no and context:
            return "{} Context: {} Log Code: {} ({})".format(msg, context, error_slug, error_url)
        if context:
            return "{} Context: {} ".format(msg, context)
        if error_no:
            return "{} Log Code: {} ({})".format(msg, error_slug, error_url)

        return msg

    def raise_config_error(self, msg, error_no, *, source_exception=None, context=None) -> "NoReturn":
        """Raise a ConfigFileError exception."""
        raise ConfigFileError(msg, error_no, self.log.name if self.log else "", context, self._url_base) \
            from source_exception

    def ignorable_runtime_exception(self, msg: str) -> None:
        """Handle ignorable runtime exception.

        During development or tests raise an exception for easier debugging. Log an error during production.
        """
        if self._debug_to_console:
            raise RuntimeError(msg)

        self.error_log(msg)

    def _logging_not_configured(self) -> "NoReturn":
        if self.machine and self.machine.is_shutting_down:
            # omit errors on shutdown
            return
        raise RuntimeError(
            "Logging has not been configured for the {} module. You must call "
            "configure_logging() before you can post a log message".
            format(self))



"""Contains the Accelerometer device."""
import math
from typing import Tuple, Optional

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.machine import MachineController
from mpf.core.platform import AccelerometerPlatform
from mpf.core.system_wide_device import SystemWideDevice
from mpf.platforms.interfaces.accelerometer_platform_interface import AccelerometerPlatformInterface


@DeviceMonitor("value")
class Accelerometer(SystemWideDevice):

    """Implements a multi-axis accelerometer.

    In modern machines, accelerometers can be used for tilt detection and to
    detect whether a machine is properly leveled.

    The accelerometer device produces a data stream of readings which MPF
    converts to g-forces, and then events can be posted when the "hit" (or
    g-force) of an accelerometer exceeds a predefined threshold.

    """

    config_section = 'accelerometers'
    collection = 'accelerometers'
    class_label = 'accelerometer'

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialise accelerometer."""
        self.platform = None        # type: Optional[AccelerometerPlatform]
        super().__init__(machine, name)

        self.history = None     # type: Optional[Tuple[float, float, float]]
        self.value = None       # type: Optional[Tuple[float, float, float]]
        self.hw_device = None   # type: Optional[AccelerometerPlatformInterface]

    async def _initialize(self):
        """Initialise and configure accelerometer."""
        await super()._initialize()
        self.platform = self.machine.get_platform_sections(
            'accelerometers', self.config['platform'])
        self.platform.assert_has_feature("accelerometers")

        if not self.platform.features['allow_empty_numbers'] and self.config['number'] is None:
            self.raise_config_error("Accelerometer must have a number.", 1)

        self.hw_device = self.platform.configure_accelerometer(self.config['number'],
                                                               self.config['platform_settings'], self)

    @classmethod
    def _calculate_vector_length(cls, x: float, y: float, z: float) -> float:
        return math.sqrt(x * x + y * y + z * z)

    # pylint: disable-msg=too-many-arguments
    def _calculate_angle(self, x1: float, y1: float, z1: float, x2: float, y2: float, z2: float) -> float:
        dividor = (self._calculate_vector_length(x1, y1, z1) *
                   self._calculate_vector_length(x2, y2, z2))

        if dividor == 0:
            return 0

        return math.acos((x1 * x2 + y1 * y2 + z1 * z2) / dividor)

    def update_acceleration(self, x: float, y: float, z: float) -> None:
        """Calculate acceleration based on readings from hardware."""
        self.value = (x, y, z)

        if not self.history:
            self.history = (x, y, z)
            dx = dy = dz = 0.0
        else:
            dx = x - self.history[0]
            dy = y - self.history[1]
            dz = z - self.history[2]

            alpha = self.config['alpha']
            self.history = (self.history[0] * alpha + x * (1 - alpha),
                            self.history[1] * alpha + y * (1 - alpha),
                            self.history[2] * alpha + z * (1 - alpha))

        self._handle_hits(dx, dy, dz)
        # only check level when we are in a stedy state
        if math.fabs(dx) + math.fabs(dy) + math.fabs(dz) < 0.05:
            self._handle_level()

    def get_level_xyz(self) -> float:
        """Return current 3D level."""
        assert self.value is not None
        return self._calculate_angle(self.config['level_x'],
                                     self.config['level_y'],
                                     self.config['level_z'],
                                     self.value[0], self.value[1],
                                     self.value[2])

    def get_level_xz(self) -> float:
        """Return current 2D x/z level."""
        assert self.value is not None
        return self._calculate_angle(self.config['level_x'],
                                     0.0, self.config['level_z'],
                                     self.value[0], 0.0, self.value[2])

    def get_level_yz(self) -> float:
        """Return current 2D y/z level."""
        assert self.value is not None
        return self._calculate_angle(0.0,
                                     self.config['level_y'],
                                     self.config['level_z'],
                                     0.0, self.value[1], self.value[2])

    def _handle_level(self) -> None:
        deviation_xyz = self.get_level_xyz()
        deviation_xz = self.get_level_xz()
        deviation_yz = self.get_level_yz()

        for max_deviation in self.config['level_limits']:
            if deviation_xyz / math.pi * 180 > max_deviation:
                self.debug_log("Deviation x: %s, y: %s, total: %s",
                               deviation_xz / math.pi * 180,
                               deviation_yz / math.pi * 180,
                               deviation_xyz / math.pi * 180)
                self.machine.events.post(
                    self.config['level_limits'][max_deviation],
                    deviation_xyz=deviation_xyz,
                    deviation_xz=deviation_xz,
                    deviation_yz=deviation_yz)

    def _handle_hits(self, dx: float, dy: float, dz: float) -> None:
        acceleration = self._calculate_vector_length(dx, dy, dz)
        for min_acceleration in self.config['hit_limits']:
            if acceleration > min_acceleration:
                self.debug_log("Received hit of %s > %s. Posting %s",
                               acceleration,
                               min_acceleration,
                               self.config['hit_limits'][min_acceleration]
                               )
                self.machine.events.post(
                    self.config['hit_limits'][min_acceleration])



from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallDeviceSwitchConfirmation(MpfTestCase):
    def __init__(self, test_map):
        super().__init__(test_map)
        self._captured = 0
        self._enter = -1
        self._missing = 0
        self._requesting = 0
        self._queue = False

    def get_config_file(self):
        return 'test_ball_device_switch_confirmation.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def _missing_ball(self, **kwargs):
        del kwargs
        self._missing += 1

    def _requesting_ball(self, balls, **kwargs):
        del kwargs
        self._requesting += balls

    def _ball_enter(self, new_balls, unclaimed_balls, **kwargs):
        del new_balls
        del kwargs
        self._enter += unclaimed_balls

    def _captured_from_pf(self, balls, **kwargs):
        del kwargs
        self._captured += balls

    def _hit_confirm(self):
        self.machine.switch_controller.process_switch("s_launcher_confirm", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_launcher_confirm", 0)
        self.advance_time_and_run(0.1)

    def test_eject_successful_to_other_trough(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device4 = self.machine.ball_devices['test_target2']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_test_target2_ball_enter',
                                        self._ball_enter)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = -1

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # request an ball
        device4.request_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self._hit_confirm()
        self.advance_time_and_run(1)

        # target2 receives and keeps ball
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.assertEqual(0, self._enter)
        self.assertEqual(-1, self._captured)

        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, self._missing)

    def test_eject_no_confirm_but_target_enter(self):
        # TODO: we may detect this problem and act accordingly (broken confirm switch)
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device4 = self.machine.ball_devices['test_target2']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_test_target2_ball_enter',
                                        self._ball_enter)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = -1
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0

        # assume we have more than one ball to prevent BallController from "fixing" pf
        self.machine.ball_controller.num_balls_known = 5

        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # request an ball
        device4.request_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        # target2 receives and keeps ball
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)
        self.assertEqual(-1, playfield.available_balls)
        self.assertEqual(-1, playfield.balls)

        # eject will fail since the eject_confirm switch was not hit
        self.advance_time_and_run(100)

        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.assertEqual(0, self._enter)
        self.assertEqual(1, self._captured)

        # no ball on pf because the pf saw an unexpected ball
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)
        self.assertEqual(1, self._missing)

        # there should be exactly 2 available_balls in the machine
        self.assertEqual(0, device1.available_balls)
        self.assertEqual(0, device2.available_balls)
        self.assertEqual(2, device4.available_balls)
        self.assertEqual(0, playfield.available_balls)

    def test_eject_successful_but_ball_never_arrives(self):
        # TODO: improve this situation
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device4 = self.machine.ball_devices['test_target2']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_test_target2_ball_enter',
                                        self._ball_enter)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0

        self.assertEqual(0, playfield.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # request an ball
        device4.request_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self._hit_confirm()
        self.advance_time_and_run(1)
        self.assertEqual(0, playfield.balls)

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        # ball never arrives
        self.advance_time_and_run(20)

        # ball should be at playfield by now and got missing
        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, self._missing)

        # since target2 != playfield. we eject a new ball to target2
        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self._hit_confirm()
        self.advance_time_and_run(1)

        # target2 captures and keeps ball
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # there should be exactly 2 available_balls in the machine
        self.assertEqual(0, device1.available_balls)
        self.assertEqual(0, device2.available_balls)
        self.assertEqual(1, device4.available_balls)
        self.assertEqual(1, playfield.available_balls)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, self._missing)

    def test_eject_successful_but_ball_never_arrives_and_drain(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device3 = self.machine.ball_devices['test_target1']
        device4 = self.machine.ball_devices['test_target2']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.assertEqual(None, self.machine.game)

        # start a game
        self.machine.switch_controller.process_switch("s_start", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)

        self.assertNotEqual(None, self.machine.game)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.assertEqual("idle", device1._state)
        self.assertEqual("ejecting", device2._state)
        self.assertEqual("waiting_for_ball", device3._state)
        self.assertEqual("idle", device4._state)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.assertEqual("idle", device1._state)
        self.assertEqual("ball_left", device2._state)
        self.assertEqual("waiting_for_ball", device3._state)
        self.assertEqual("idle", device4._state)

        self._hit_confirm()
        self.advance_time_and_run(1)
        self.assertEqual(0, playfield.balls)

        # ball never arrives and goes to pf. its not yet missing
        self.advance_time_and_run(10)
        self.assertEqual(0, self._missing)

        # ball drains
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, device1.balls)
        self.assertEqual(None, self.machine.game)

        # ball should not be at playfield by now
        self.assertEqual(-1, playfield.balls)
        self.assertEqual(1, self._captured)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # ball device notices that ball went missing
        self.advance_time_and_run(100)

        self.assertEqual("idle", device1._state)
        self.assertEqual("idle", device2._state)
        self.assertEqual("idle", device3._state)
        self.assertEqual("idle", device4._state)

        # there should be exactly 2 available_balls in the machine
        self.assertEqual(2, device1.available_balls)
        self.assertEqual(0, device2.available_balls)
        self.assertEqual(0, device3.available_balls)
        self.assertEqual(0, device4.available_balls)
        self.assertEqual(0, playfield.available_balls)

        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._missing)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

    def test_eject_successful_but_ball_never_arrives_and_stays_on_pf(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device3 = self.machine.ball_devices['test_target1']
        device4 = self.machine.ball_devices['test_target2']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.assertEqual(None, self.machine.game)

        # start a game
        self.machine.switch_controller.process_switch("s_start", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)

        self.assertNotEqual(None, self.machine.game)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)
        self.assertEqual(1, self.machine.ball_controller.num_balls_known)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.assertEqual("idle", device1._state)
        self.assertEqual("ejecting", device2._state)
        self.assertEqual("waiting_for_ball", device3._state)
        self.assertEqual("idle", device4._state)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.assertEqual("idle", device1._state)
        self.assertEqual("ball_left", device2._state)
        self.assertEqual("waiting_for_ball", device3._state)
        self.assertEqual("idle", device4._state)

        self._hit_confirm()
        self.advance_time_and_run(1)
        self.assertEqual(0, playfield.balls)

        # ball never arrives and goes to pf. its not yet missing
        self.advance_time_and_run(10)
        self.assertEqual(0, self._missing)

        # ball should not be at playfield by now
        self.assertEqual(0, playfield.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # ball device notices that ball went missing
        self.advance_time_and_run(100)
        self.assertEqual(1, self._missing)

        self.assertEqual("idle", device1._state)
        self.assertEqual("idle", device2._state)
        self.assertEqual("idle", device3._state)
        self.assertEqual("idle", device4._state)

        # there should be exactly 1 available_ball in the machine
        self.assertEqual(0, device1.available_balls)
        self.assertEqual(0, device2.available_balls)
        self.assertEqual(0, device3.available_balls)
        self.assertEqual(0, device4.available_balls)
        self.assertEqual(1, playfield.available_balls)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, self._missing)
        self.assertEqual(0, self._captured)

        self.assertEqual(1, self.machine.ball_controller.num_balls_known)

    def test_ball_return_in_launcher(self):
        # add two initial balls to trough
        self.hit_switch_and_run("s_ball_switch1", 1)
        self.hit_switch_and_run("s_ball_switch2", 1)
        self.assertEqual(2, self.machine.ball_devices["test_trough"].balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertBallsOnPlayfield(0, "playfield")
        self.assertAvailableBallsOnPlayfield(0, "playfield")

        self.hit_switch_and_run("s_ball_switch_target1", 1)

        self.release_switch_and_run("s_ball_switch_target1", 1)
        self.assertBallsOnPlayfield(0, "playfield")
        self.assertAvailableBallsOnPlayfield(1, "playfield")

        self.machine.ball_devices["test_target1"].request_ball()
        self.assertEqual(1, self.machine.ball_devices["test_target1"].available_balls)

        self.advance_time_and_run()
        self.assertEqual("pulsed_10", self.machine.coils["eject_coil1"].hw_driver.state)
        self.release_switch_and_run("s_ball_switch2", 1)
        self.advance_time_and_run(8)

        self.hit_switch_and_run("s_ball_switch_launcher", 1)
        # launcher should eject
        self.advance_time_and_run(1)
        self.assertEqual("pulsed_10", self.machine.coils["eject_coil2"].hw_driver.state)
        self.machine.coils["eject_coil2"].hw_driver.state = None

        # it leaves the switch
        self.release_switch_and_run("s_ball_switch_launcher", 3)

        # switch goes active again
        self.hit_switch_and_run("s_ball_switch_launcher", 1)
        self.assertEqual(None, self.machine.coils["eject_coil2"].hw_driver.state)
        self.assertEqual(1, self.machine.ball_devices["test_target1"].available_balls)

        self.advance_time_and_run(9)
        self.assertEqual("pulsed_10", self.machine.coils["eject_coil2"].hw_driver.state)
        self.assertEqual(1, self.machine.ball_devices["test_target1"].available_balls)

        self.release_switch_and_run("s_ball_switch_launcher", 1)

        self._hit_confirm()
        self.advance_time_and_run()

        self.hit_switch_and_run("s_ball_switch_target1", 1)

        self.assertBallsOnPlayfield(1, "playfield")
        self.assertAvailableBallsOnPlayfield(1, "playfield")



"""Controls the rules on all platforms."""
from collections import namedtuple

from typing import Optional

from mpf.core.mpf_controller import MpfController
from mpf.core.platform import DriverPlatform, SwitchSettings, DriverSettings, RepulseSettings
from mpf.core.switch_controller import SwitchHandler
from mpf.devices.driver import Driver
from mpf.devices.switch import Switch
from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import

SwitchRuleSettings = namedtuple("SwitchRuleSettings", ["switch", "invert", "debounce"])
DriverRuleSettings = namedtuple("DriverRuleSettings", ["driver", "recycle"])
PulseRuleSettings = namedtuple("PulseRuleSettings", ["power", "duration"])
EosRuleSettings = namedtuple("RepulseRuleSettings", ["enable_repulse", "debounce_ms"])
HoldRuleSettings = namedtuple("HoldRuleSettings", ["power"])
HardwareRule = namedtuple("HardwareRule", ["platform", "switch_settings", "driver_settings", "switch_key",
                                           "software_rule_handler"])


class SoftwareEosRepulseManager:

    """Manager to implement eos repulse in software."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, machine: "MachineController", enable_switch: SwitchRuleSettings, eos_switch: SwitchRuleSettings,
                 driver: DriverSettings, repulse_settings: EosRuleSettings):
        """Initialize software eos repulse manager."""
        self.machine = machine
        self.enable_switch = enable_switch
        self.eos_switch = eos_switch
        self.driver = driver
        self.repulse_settings = repulse_settings
        self._button_is_active = False
        self._is_eos_closed_long_enough = False
        self._handlers = []

        self._handlers.append(self.machine.switch_controller.add_switch_handler_obj(
            enable_switch.switch,
            state=0 if enable_switch.invert else 1,
            callback=self._button_active))
        self._handlers.append(self.machine.switch_controller.add_switch_handler_obj(
            enable_switch.switch,
            state=1 if enable_switch.invert else 0,
            callback=self._button_inactive))
        self._handlers.append(self.machine.switch_controller.add_switch_handler_obj(
            eos_switch.switch,
            state=0 if eos_switch.invert else 1,
            callback=self._eos_closed_long_enough,
            ms=self.repulse_settings.debounce_ms))
        self._handlers.append(self.machine.switch_controller.add_switch_handler_obj(
            eos_switch.switch,
            state=1 if eos_switch.invert else 0,
            callback=self._repulse_on_eos_open))

    def stop(self):
        """Stop software repulse."""
        self.machine.switch_controller.remove_switch_handler_by_keys(self._handlers)

    def _button_active(self, **kwargs):
        del kwargs
        self._button_is_active = True

    def _button_inactive(self, **kwargs):
        del kwargs
        self._button_is_active = False
        self.driver.hw_driver.disable()

    def _eos_closed_long_enough(self, **kwargs):
        del kwargs
        self._is_eos_closed_long_enough = True

    def _repulse_on_eos_open(self, **kwargs):
        del kwargs
        if not self._button_is_active or not self._is_eos_closed_long_enough:
            return

        self._is_eos_closed_long_enough = False

        if self.driver.hold_settings:
            self.driver.hw_driver.enable(self.driver.pulse_settings, self.driver.hold_settings)
        else:
            self.driver.hw_driver.pulse(self.driver.pulse_settings)


class PlatformController(MpfController):

    """Manages all platforms and rules."""

    config_name = "platform_controller"

    @staticmethod
    def _check_and_get_platform(switch: Switch, driver: Driver) -> DriverPlatform:
        assert driver.platform is not None
        assert switch.platform is not None
        if driver.platform != switch.platform:
            raise AssertionError("Switch and Coil have to use the same platform")

        return driver.platform

    def _setup_switch_callback_for_psu(self, switch: Switch, driver: Driver, switch_settings: SwitchSettings,
                                       driver_settings: DriverSettings) -> Optional[SwitchHandler]:
        """Set up a switch handler which informs the PSU about pulses performed by the rule."""
        if driver_settings.pulse_settings.duration == 0:
            return None

        key = self.machine.switch_controller.add_switch_handler(
            switch_name=switch.name,
            state=0 if switch_settings.invert else 1,
            callback=self._notify_psu_about_pulse,
            callback_kwargs={"driver": driver,
                             "pulse_ms": driver_settings.pulse_settings.duration})

        return key

    @staticmethod
    def _notify_psu_about_pulse(driver: Driver, pulse_ms: int):
        """Notify PSU that a pulse via a rule happened."""
        driver.config['psu'].notify_about_instant_pulse(pulse_ms=pulse_ms)

    # pylint: disable-msg=too-many-arguments
    def _get_repulse_settings(self, eos_settings: Optional[EosRuleSettings], enable_switch, eos_switch,
                              driver_settings, platform) -> Optional[RepulseSettings]:
        """Return repulse settings for rule."""
        if eos_settings:
            repulse_settings = RepulseSettings(
                enable_repulse=eos_settings.enable_repulse,
                debounce_ms=eos_settings.debounce_ms
            )
            if repulse_settings and repulse_settings.enable_repulse and not platform.features["hardware_eos_repulse"]:
                # Platform does not support EOS repulse in hardware -> emulate it in software
                software_eos_handler = SoftwareEosRepulseManager(self.machine, enable_switch, eos_switch,
                                                                 driver_settings, repulse_settings)
                return None, software_eos_handler

            return repulse_settings, None

        return None, None

    @staticmethod
    def _get_configured_switch(switch: SwitchRuleSettings) -> SwitchSettings:
        """Return configured switch for rule."""
        return SwitchSettings(
            hw_switch=switch.switch.hw_switch,
            invert=switch.invert != switch.switch.invert,
            debounce=switch.debounce)

    @staticmethod
    def _get_configured_driver_with_hold(driver: DriverRuleSettings, pulse_setting: Optional[PulseRuleSettings],
                                         hold_settings: Optional[HoldRuleSettings]) -> DriverSettings:
        """Return configured driver with hold > 0 for rule."""
        pulse_duration = driver.driver.get_and_verify_pulse_ms(pulse_setting.duration if pulse_setting else None)
        pulse_power = driver.driver.get_and_verify_pulse_power(pulse_setting.power if pulse_setting else None)
        hold_power = driver.driver.get_and_verify_hold_power(hold_settings.power if hold_settings else None)

        if hold_power == 0.0:
            raise AssertionError("Cannot enable driver with hold_power 0.0")

        return DriverSettings(
            hw_driver=driver.driver.hw_driver,
            pulse_settings=PulseSettings(duration=pulse_duration, power=pulse_power),
            hold_settings=HoldSettings(power=hold_power),
            recycle=driver.recycle)

    @staticmethod
    def _get_configured_driver_no_hold(driver: DriverRuleSettings,
                                       pulse_setting: Optional[PulseRuleSettings]) -> DriverSettings:
        """Return configured driver without hold for rule."""
        pulse_duration = driver.driver.get_and_verify_pulse_ms(pulse_setting.duration if pulse_setting else None)
        pulse_power = driver.driver.get_and_verify_pulse_power(pulse_setting.power if pulse_setting else None)
        return DriverSettings(
            hw_driver=driver.driver.hw_driver,
            pulse_settings=PulseSettings(duration=pulse_duration, power=pulse_power),
            hold_settings=None,
            recycle=driver.recycle)

    def set_pulse_on_hit_rule(self, enable_switch: SwitchRuleSettings,
                              driver: DriverRuleSettings,
                              pulse_setting: PulseRuleSettings = None) -> HardwareRule:
        """Add pulse on hit rule to driver.

        Always do the full pulse. Even when the switch is released.

        Args:
        ----
            enable_switch: Switch which triggers the rule.
            driver: .. class:: DriverRuleSettings
            pulse_setting: .. class:: PulseRuleSettings
        """
        platform = self._check_and_get_platform(enable_switch.switch, driver.driver)

        enable_settings = self._get_configured_switch(enable_switch)
        driver_settings = self._get_configured_driver_no_hold(driver, pulse_setting)

        platform.set_pulse_on_hit_rule(enable_settings, driver_settings)

        switch_key = self._setup_switch_callback_for_psu(enable_switch.switch, driver.driver, enable_settings,
                                                         driver_settings)

        self.machine.bcp.interface.send_driver_event(
            action="pulse_on_hit",
            enable_switch_number=enable_switch.switch.hw_switch.number,
            enable_switch_name=enable_switch.switch.name,
            enable_switch_invert=enable_settings.invert,
            enable_switch_debounce=enable_settings.debounce,
            coil_number=driver.driver.hw_driver.number,
            coil_name=driver.driver.name,
            coil_pulse_power=driver_settings.pulse_settings.power,
            coil_pulse_ms=driver_settings.pulse_settings.duration,
            coil_hold_power=0,
            coil_recycle=driver_settings.recycle)

        return HardwareRule(platform=platform, switch_settings=[enable_settings], driver_settings=driver_settings,
                            switch_key=switch_key, software_rule_handler=None)

    def set_delayed_pulse_on_hit_rule(self, enable_switch: SwitchRuleSettings,
                                      driver: DriverRuleSettings,
                                      delay_ms, pulse_setting: PulseRuleSettings = None) -> HardwareRule:
        """Add delayed pulse on hit rule to driver.

        Always do the full pulse. Even when the switch is released. Pulse is delayed accurately by the hardware.

        Args:
        ----
            enable_switch: Switch which triggers the rule.
            driver: .. class:: DriverRuleSettings
            delay_ms: Delay before the pulse in ms
            pulse_setting: .. class:: PulseRuleSettings
        """
        platform = self._check_and_get_platform(enable_switch.switch, driver.driver)

        enable_settings = self._get_configured_switch(enable_switch)
        driver_settings = self._get_configured_driver_no_hold(driver, pulse_setting)

        platform.set_delayed_pulse_on_hit_rule(enable_settings, driver_settings, delay_ms)

        switch_key = self._setup_switch_callback_for_psu(enable_switch.switch, driver.driver, enable_settings,
                                                         driver_settings)

        self.machine.bcp.interface.send_driver_event(
            action="delayed_pulse_on_hit",
            enable_switch_number=enable_switch.switch.hw_switch.number,
            enable_switch_name=enable_switch.switch.name,
            enable_switch_invert=enable_settings.invert,
            enable_switch_debounce=enable_settings.debounce,
            coil_number=driver.driver.hw_driver.number,
            coil_name=driver.driver.name,
            coil_pulse_power=driver_settings.pulse_settings.power,
            coil_pulse_ms=driver_settings.pulse_settings.duration,
            coil_hold_power=0,
            coil_recycle=driver_settings.recycle,
            delay_ms=delay_ms)

        return HardwareRule(platform=platform, switch_settings=[enable_settings], driver_settings=driver_settings,
                            switch_key=switch_key, software_rule_handler=None)

    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchRuleSettings,
                                          driver: DriverRuleSettings,
                                          pulse_setting: PulseRuleSettings = None) -> HardwareRule:
        """Add pulse on hit and release rule to driver.

        Pulse a driver but cancel pulse when switch is released.

        Args:
        ----
            enable_switch: Switch which triggers the rule.
            driver: .. class:: DriverRuleSettings
            pulse_setting: .. class:: PulseRuleSettings
        """
        platform = self._check_and_get_platform(enable_switch.switch, driver.driver)

        enable_settings = self._get_configured_switch(enable_switch)
        driver_settings = self._get_configured_driver_no_hold(driver, pulse_setting)

        platform.set_pulse_on_hit_and_release_rule(enable_settings, driver_settings)

        switch_key = self._setup_switch_callback_for_psu(enable_switch.switch, driver.driver, enable_settings,
                                                         driver_settings)

        self.machine.bcp.interface.send_driver_event(
            action="pulse_on_hit_and_release",
            enable_switch_number=enable_switch.switch.hw_switch.number,
            enable_switch_name=enable_switch.switch.name,
            enable_switch_invert=enable_settings.invert,
            enable_switch_debounce=enable_settings.debounce,
            coil_number=driver.driver.hw_driver.number,
            coil_name=driver.driver.name,
            coil_pulse_power=driver_settings.pulse_settings.power,
            coil_pulse_ms=driver_settings.pulse_settings.duration,
            coil_hold_power=0,
            coil_recycle=driver_settings.recycle)

        return HardwareRule(platform=platform, switch_settings=[enable_settings], driver_settings=driver_settings,
                            switch_key=switch_key, software_rule_handler=None)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchRuleSettings,
                                                     driver: DriverRuleSettings,
                                                     pulse_setting: PulseRuleSettings = None,
                                                     hold_settings: HoldRuleSettings = None) -> HardwareRule:
        """Add pulse on hit and enable and release rule to driver.

        Pulse and enable a driver. Cancel pulse and enable if switch is released.

        Args:
        ----
            enable_switch: Switch which triggers the rule.
            driver: Driver to trigger.
            pulse_setting: .. class:: PulseRuleSettings
            hold_settings: .. class:: HoldRuleSettings
        """
        platform = self._check_and_get_platform(enable_switch.switch, driver.driver)

        enable_settings = self._get_configured_switch(enable_switch)
        driver_settings = self._get_configured_driver_with_hold(driver, pulse_setting, hold_settings)

        platform.set_pulse_on_hit_and_enable_and_release_rule(enable_settings, driver_settings)

        switch_key = self._setup_switch_callback_for_psu(enable_switch.switch, driver.driver, enable_settings,
                                                         driver_settings)

        self.machine.bcp.interface.send_driver_event(
            action="pulse_on_hit_and_enable_and_release",
            enable_switch_number=enable_switch.switch.hw_switch.number,
            enable_switch_name=enable_switch.switch.name,
            enable_switch_invert=enable_settings.invert,
            enable_switch_debounce=enable_settings.debounce,
            coil_number=driver.driver.hw_driver.number,
            coil_name=driver.driver.name,
            coil_pulse_power=driver_settings.pulse_settings.power,
            coil_pulse_ms=driver_settings.pulse_settings.duration,
            coil_hold_power=driver_settings.hold_settings,
            coil_recycle=driver_settings.recycle)

        return HardwareRule(platform=platform, switch_settings=[enable_settings], driver_settings=driver_settings,
                            switch_key=switch_key, software_rule_handler=None)

    # pylint: disable-msg=too-many-arguments
    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchRuleSettings,
                                                      eos_switch: SwitchRuleSettings,
                                                      driver: DriverRuleSettings,
                                                      pulse_setting: Optional[PulseRuleSettings] = None,
                                                      eos_settings: Optional[EosRuleSettings] = None
                                                      ) -> HardwareRule:
        """Add pulse on hit and release and disable rule to driver.

        Pulse driver. Cancel pulse when switch is released or a disable switch is hit.

        Args:
        ----
            enable_switch: Switch to enable coil.
            eos_switch: Switch to cancel pulse and trigger repulses.
            driver: Driver to enable.
            pulse_setting: Pulse settings.
            eos_settings: How to repulse the coil when EOS opens.
        """
        platform = self._check_and_get_platform(enable_switch.switch, driver.driver)
        self._check_and_get_platform(eos_switch.switch, driver.driver)

        enable_settings = self._get_configured_switch(enable_switch)
        disable_settings = self._get_configured_switch(eos_switch)
        driver_settings = self._get_configured_driver_no_hold(driver, pulse_setting)
        repulse_settings, software_eos_handler = self._get_repulse_settings(
            eos_settings, enable_switch, eos_switch, driver_settings, platform)

        platform.set_pulse_on_hit_and_release_and_disable_rule(
            enable_settings, disable_settings, driver_settings, repulse_settings)

        switch_key = self._setup_switch_callback_for_psu(enable_switch.switch, driver.driver, enable_settings,
                                                         driver_settings)

        self.machine.bcp.interface.send_driver_event(
            action="pulse_on_hit_and_enable_and_release_and_disable",
            enable_switch_number=enable_switch.switch.hw_switch.number,
            enable_switch_name=enable_switch.switch.name,
            enable_switch_invert=enable_settings.invert,
            enable_switch_debounce=enable_settings.debounce,
            disable_switch_number=eos_switch.switch.hw_switch.number,
            disable_switch_name=eos_switch.switch.name,
            disable_switch_invert=disable_settings.invert,
            disable_switch_debounce=disable_settings.debounce,
            coil_number=driver.driver.hw_driver.number,
            coil_name=driver.driver.name,
            coil_pulse_power=driver_settings.pulse_settings.power,
            coil_pulse_ms=driver_settings.pulse_settings.duration,
            coil_hold_power=driver_settings.hold_settings,
            coil_recycle=driver_settings.recycle)

        return HardwareRule(platform=platform, switch_settings=[enable_settings, disable_settings],
                            driver_settings=driver_settings, switch_key=switch_key,
                            software_rule_handler=software_eos_handler)

    # pylint: disable-msg=too-many-arguments
    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchRuleSettings,
                                                                 eos_switch: SwitchRuleSettings,
                                                                 driver: DriverRuleSettings,
                                                                 pulse_setting: Optional[PulseRuleSettings] = None,
                                                                 hold_settings: Optional[HoldRuleSettings] = None,
                                                                 eos_settings: Optional[EosRuleSettings] = None
                                                                 ) -> HardwareRule:
        """Add pulse on hit and enable and release and disable rule to driver.

        Pulse and then enable driver. Cancel pulse and enable when switch is released or a disable switch is hit.

        Args:
        ----
            enable_switch: Switch to enable coil.
            eos_switch: Switch to switch from pulse to hold and to trigger repulses.
            driver: Driver to enable.
            pulse_setting: Pulse settings.
            hold_settings: Hold settings.
            eos_settings: How to repulse the coil when EOS opens.
        """
        platform = self._check_and_get_platform(enable_switch.switch, driver.driver)
        self._check_and_get_platform(eos_switch.switch, driver.driver)

        enable_settings = self._get_configured_switch(enable_switch)
        disable_settings = self._get_configured_switch(eos_switch)
        driver_settings = self._get_configured_driver_with_hold(driver, pulse_setting, hold_settings)
        repulse_settings, software_eos_handler = self._get_repulse_settings(
            eos_settings, enable_switch, eos_switch, driver_settings, platform)

        platform.set_pulse_on_hit_and_enable_and_release_and_disable_rule(
            enable_settings, disable_settings, driver_settings, repulse_settings)

        switch_key = self._setup_switch_callback_for_psu(enable_switch.switch, driver.driver, enable_settings,
                                                         driver_settings)

        self.machine.bcp.interface.send_driver_event(
            action="pulse_on_hit_and_enable_and_release_and_disable",
            enable_switch_number=enable_switch.switch.hw_switch.number,
            enable_switch_name=enable_switch.switch.name,
            enable_switch_invert=enable_settings.invert,
            enable_switch_debounce=enable_settings.debounce,
            disable_switch_number=eos_switch.switch.hw_switch.number,
            disable_switch_name=eos_switch.switch.name,
            disable_switch_invert=disable_settings.invert,
            disable_switch_debounce=disable_settings.debounce,
            coil_number=driver.driver.hw_driver.number,
            coil_name=driver.driver.name,
            coil_pulse_power=driver_settings.pulse_settings.power,
            coil_pulse_ms=driver_settings.pulse_settings.duration,
            coil_hold_power=driver_settings.hold_settings,
            coil_recycle=driver_settings.recycle)

        return HardwareRule(platform=platform, switch_settings=[enable_settings, disable_settings],
                            driver_settings=driver_settings, switch_key=switch_key,
                            software_rule_handler=software_eos_handler)

    def clear_hw_rule(self, rule: HardwareRule):
        """Clear all rules for switch and this driver.

        Args:
        ----
            rule: Hardware rule to clean.
        """
        for switch_settings in rule.switch_settings:
            rule.platform.clear_hw_rule(switch_settings, rule.driver_settings)

            self.machine.bcp.interface.send_driver_event(
                action="remove",
                enable_switch_number=switch_settings.hw_switch.number,
                enable_switch_invert=switch_settings.invert,
                coil_number=rule.driver_settings.hw_driver.number)

        if rule.switch_key:
            self.machine.switch_controller.remove_switch_handler_by_key(rule.switch_key)

        if rule.software_rule_handler:
            rule.software_rule_handler.stop()



"""Contains code for PIN2DMD."""
import logging

import threading
from mpf.core.utility_functions import Util

from mpf.platforms.interfaces.dmd_platform import DmdPlatformInterface
from mpf.core.platform import RgbDmdPlatform

# pylint: disable-msg=ungrouped-imports
try:
    import usb.core
except ImportError as e:
    IMPORT_FAILED = e
else:
    IMPORT_FAILED = None    # type: ignore


GAMMA_TABLE = [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
               2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3,
               3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5,
               5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7,
               7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10,
               11, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 14, 14,
               14, 14, 15, 15, 15, 16, 16, 16, 16, 17, 17, 17, 18, 18, 18, 18,
               19, 19, 19, 20, 20, 20, 21, 21, 21, 22, 22, 22, 23, 23, 23, 24,
               24, 24, 25, 25, 25, 26, 26, 27, 27, 27, 28, 28, 29, 29, 29, 30,
               30, 31, 31, 31, 32, 32, 33, 33, 34, 34, 35, 35, 35, 36, 36, 37,
               37, 38, 38, 39, 39, 40, 40, 41, 41, 42, 42, 43, 43, 44, 44, 45,
               45, 46, 47, 47, 48, 48, 49, 49, 50, 50, 51, 52, 52, 53, 53, 54,
               55, 55, 56, 56, 57, 58, 58, 59, 60, 60, 61, 62, 62, 63, 63, 63]


class Pin2DmdHardwarePlatform(RgbDmdPlatform):

    """PIN2DMD RGB DMD hardware."""

    __slots__ = ["device", "config"]

    def __init__(self, machine):
        """Initialise PIN2DMD."""
        super().__init__(machine)
        self.features['tickless'] = True
        self.config = self.machine.config_validator.validate_config("pin2dmd", self.machine.config.get('pin2dmd', {}))
        self._configure_device_logging_and_debug('PIN2DMD', self.config)
        self.log.debug("Configuring PIN2DMD hardware interface.")
        self.device = Pin2DmdDevice(machine, self.debug, self.config['resolution'], self.config['panel'])

        if IMPORT_FAILED:
            raise AssertionError('Failed to load pyusb. Did you install pyusb? '
                                 'Try: "pip3 install pyusb".') from IMPORT_FAILED

    async def initialize(self):
        """Initialise platform."""
        await self.device.connect()

    def stop(self):
        """Stop platform."""
        self.device.stop()
        self.device = None

    def __repr__(self):
        """Return string representation."""
        return '<Platform.Pin2Dmd>'

    def configure_rgb_dmd(self, name: str):
        """Configure rgb dmd."""
        if name != "default":
            self.raise_config_error("Use dmd name 'default' for PIN2DMD.", 1)
        return self.device


class Pin2DmdDevice(DmdPlatformInterface):

    """A PIN2DMD device."""

    __slots__ = ["writer", "current_frame", "new_frame_event", "machine", "log", "device", "brightness",
                 "debug", "resolution", "panel"]

    def __init__(self, machine, debug, resolution, panel):
        """Initialise smart matrix device."""
        self.writer = None
        self.current_frame = None
        self.new_frame_event = None
        self.machine = machine
        self.device = None
        self.brightness = 255
        self.log = logging.getLogger('Pin2DmdDevice')
        self.debug = debug
        self.resolution = resolution
        self.panel = panel

    def _send_brightness(self, brightness):
        data = [0x00] * 2052
        data[0] = 0x81
        data[1] = 0xc3
        data[2] = 0xe7
        data[3] = 0xff
        data[4] = 0x08
        data[17] = brightness

        if self.debug:
            self.log.debug("Writing 0x01, %s", "".join(" 0x%02x" % b for b in data))

        self.device.write(0x01, data)

    def _send_frame(self, buffer):
        if self.resolution == "128x32":
            elements = 2048
        else:
            elements = 6144

        output_buffer = [0] * (elements * 6 + 4)

        output_buffer[0] = 0x81
        output_buffer[1] = 0xC3
        output_buffer[2] = 0xE9
        output_buffer[3] = 18

        for i in range(0, elements):
            idx = i * 3

            if self.panel == "rgb":
                # use these mappings for RGB panels
                pixel_r = buffer[idx]
                pixel_g = buffer[idx + 1]
                pixel_b = buffer[idx + 2]
                # lower half of display
                pixel_rl = buffer[elements * 3 + idx]
                pixel_gl = buffer[elements * 3 + idx + 1]
                pixel_bl = buffer[elements * 3 + idx + 2]
            else:
                # use these mappings for RBG panels
                pixel_r = buffer[idx]
                pixel_g = buffer[idx + 2]
                pixel_b = buffer[idx + 1]
                # lower half of display
                pixel_rl = buffer[elements * 3 + idx]
                pixel_gl = buffer[elements * 3 + idx + 2]
                pixel_bl = buffer[elements * 3 + idx + 1]

            # color correction
            pixel_r = GAMMA_TABLE[pixel_r]
            pixel_g = GAMMA_TABLE[pixel_g]
            pixel_b = GAMMA_TABLE[pixel_b]

            pixel_rl = GAMMA_TABLE[pixel_rl]
            pixel_gl = GAMMA_TABLE[pixel_gl]
            pixel_bl = GAMMA_TABLE[pixel_bl]

            target_idx = i + 4

            for _ in range(0, 6):
                output_buffer[target_idx] = ((pixel_gl & 1) << 5) | ((pixel_bl & 1) << 4) | ((pixel_rl & 1) << 3) |\
                                            ((pixel_g & 1) << 2) | ((pixel_b & 1) << 1) | ((pixel_r & 1) << 0)
                pixel_r >>= 1
                pixel_g >>= 1
                pixel_b >>= 1
                pixel_rl >>= 1
                pixel_gl >>= 1
                pixel_bl >>= 1
                target_idx += elements

        if self.debug:
            self.log.debug("Writing 0x01, %s, 1000", "".join(" 0x%02x" % b for b in output_buffer))

        self.device.write(0x01, output_buffer, 1000)

    def _feed_hardware(self):
        """Feed hardware in separate thread.

        Wait for new_frame_event and send the last frame. If no event happened for 1s refresh the last frame.
        """
        current_brightness = None
        while not self.machine.thread_stopper.is_set():
            # wait for new frame or timeout
            self.new_frame_event.wait(1)

            # clear event
            self.new_frame_event.clear()

            # set brightness if it changed
            if self.brightness != current_brightness:
                current_brightness = self.brightness
                self._send_brightness(current_brightness)

            # do not crash on missing frame
            if self.current_frame is None:
                continue

            # send frame
            self._send_frame(self.current_frame)

    async def connect(self):
        """Connect to Pin2Dmd device."""
        self.log.info("Connecting to Pin2DMD RGB DMD")
        self.device = usb.core.find(idVendor=0x0314, idProduct=0xE457)
        if self.device is None:
            raise AssertionError('Pin2Dmd USB device not found')

        self.new_frame_event = threading.Event()
        self.writer = self.machine.clock.loop.run_in_executor(None, self._feed_hardware)
        self.writer.add_done_callback(Util.raise_exceptions)

        self.log.info("Connected to Pin2DMD")

    def set_brightness(self, brightness: float):
        """Set brightness."""
        if brightness < 0.0 or brightness > 1.0:
            raise AssertionError("Brightness has to be between 0 and 1.")
        self.brightness = int(brightness * 255)

    def stop(self):
        """Stop platform."""

    def update(self, data):
        """Update DMD data."""
        self.current_frame = bytearray(data)
        self.new_frame_event.set()



"""Test coils."""
from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings

from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestDeviceDriver(MpfTestCase):

    def get_config_file(self):
        return 'coils.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/device/'

    def get_platform(self):
        return 'smart_virtual'

    def testBasicFunctions(self):
        # Make sure hardware devices have been configured for tests
        self.assertIn('coil_01', self.machine.coils)
        self.assertIn('coil_02', self.machine.coils)

        # Setup platform function mock to test coil
        self.machine.coils["coil_01"].hw_driver.disable = MagicMock()
        self.machine.coils["coil_01"].hw_driver.enable = MagicMock()
        self.machine.coils["coil_01"].hw_driver.pulse = MagicMock()

        self.machine.coils["coil_01"].enable()
        self.machine.coils["coil_01"].hw_driver.enable.assert_called_with(PulseSettings(power=1.0, duration=30),
                                                                       HoldSettings(power=1.0))
        self.machine.coils["coil_01"].pulse(100)
        self.machine.coils["coil_01"].hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=100))
        self.machine.coils["coil_01"].disable()
        self.machine.coils["coil_01"].hw_driver.disable.assert_called_with()

        self.machine.coils["coil_03"].hw_driver.disable = MagicMock()
        self.machine.coils["coil_03"].hw_driver.enable = MagicMock()
        self.machine.coils["coil_03"].hw_driver.pulse = MagicMock()

        # test default pulse_ms
        self.machine.config['mpf']['default_pulse_ms'] = 23
        self.machine.coils["coil_03"].pulse()
        self.machine.coils["coil_03"].hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=23))

        # test power
        self.machine.config['mpf']['default_pulse_ms'] = 40
        self.machine.coils["coil_03"].pulse(pulse_power=1.0)
        self.machine.coils["coil_03"].hw_driver.pulse.assert_called_with(PulseSettings(power=1.0, duration=40))

        self.machine.coils["coil_03"].pulse(pulse_power=0.5)
        self.machine.coils["coil_03"].hw_driver.pulse.assert_called_with(PulseSettings(power=0.5, duration=40))

        self.machine.coils["coil_01"].enable(pulse_power=0.7, hold_power=0.3)
        self.machine.coils["coil_01"].hw_driver.enable.assert_called_with(PulseSettings(power=0.7, duration=30),
                                                                       HoldSettings(power=0.3))

        # test long pulse with delay
        self.machine.coils["coil_03"].hw_driver.pulse = MagicMock()
        self.machine.coils["coil_03"].hw_driver.enable = MagicMock()
        self.machine.coils["coil_03"].hw_driver.disable = MagicMock()
        self.machine.coils["coil_03"].pulse(pulse_ms=500)
        self.machine.coils["coil_03"].hw_driver.enable.assert_called_with(PulseSettings(power=1.0, duration=0),
                                                                       HoldSettings(power=1.0))
        self.machine.coils["coil_03"].hw_driver.pulse.assert_not_called()
        self.advance_time_and_run(.5)

        self.machine.coils["coil_03"].hw_driver.disable.assert_called_with()

    def testMaxHoldDuration(self):
        coil = self.machine.coils["coil_max_hold_duration"]

        # check that coil disables after max_hold_duration (5s)
        coil.enable()
        self.advance_time_and_run(.5)
        self.assertEqual("enabled", coil.hw_driver.state)
        self.advance_time_and_run(4)
        self.assertEqual("enabled", coil.hw_driver.state)
        self.advance_time_and_run(1)
        self.assertEqual("disabled", coil.hw_driver.state)

        # make sure a disable resets the timer
        coil.enable()
        self.advance_time_and_run(3.0)
        self.assertEqual("enabled", coil.hw_driver.state)
        coil.disable()
        self.advance_time_and_run(.5)
        self.assertEqual("disabled", coil.hw_driver.state)
        coil.enable()
        self.advance_time_and_run(3.0)
        self.assertEqual("enabled", coil.hw_driver.state)



"""Deprecated command to start MPF game engine."""

import sys


class Command:  # pragma: no cover

    """Deprecated command."""

    def __init__(self, mpf_path, machine_path, args):
        """Show error."""
        del mpf_path
        del machine_path
        del args
        print('"mpf core" has been renamed to "mpf game", so run that instead.')
        sys.exit(-1)
